import os
import sys
from sphinx.ext.autodoc import importer

sys.path.insert(0, os.path.abspath('..'))

project = 'sudio'
copyright = '2024, mrzahaki'
author = 'Hossein Zahak'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'breathe',
    'exhale'
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
autodoc_member_order = 'bysource'
autoclass_content = 'both'

breathe_projects = {
    "sudio": "./doxyoutput/xml"
}
breathe_default_project = "sudio"

exhale_args = {
    "containmentFolder":     "./api",
    "rootFileName":          "io_root.rst",
    "doxygenStripFromPath":  '..',
    "rootFileTitle":         "API",
    "createTreeView":        True,
    "exhaleExecutesDoxygen": True,
    "exhaleDoxygenStdin":    """
        INPUT                  = ../sudio/io ../sudio/rateshift
        EXCLUDE_PATTERNS       = */miniaudio/* */dr_libs/* */portaudio/* */libsamplerate/* */*.py */libmp3lame-CMAKE/* */lame/* */flac/* */ogg/* */vorbis/*
        FILTER_PATTERNS        = *.cpp=python doxyfilter.py
        RECURSIVE              = YES 
        GENERATE_XML           = YES
        GENERATE_HTML          = NO
        XML_OUTPUT             = xml
        INCLUDE_PATH           = ../cache/pybind11
    """,
}
breathe_implementation_filename_extensions = ['.c', '.cc', '.cpp', '.hpp']
primary_domain = 'cpp'
highlight_language = 'cpp'
html_show_sourcelink = False
autodoc_mock_imports = [
    'sudio._suio', 
    'sudio.io', 
    'scipy', 
    'numpy',
    'sudio.process.fx._tempo',
    'sudio.process.fx._fade_envelope',
    'sudio.process.fx._channel_mixer',
    "sudio.process.fx._pitch_shifter",
    "sudio.utils.math",
    'sudio._rateshift',
    'sudio.rateshift',
    ]
html_context = {
    "google_analytics_id": "G-RLP20V08DB",
}
html_css_files = [
    'style.css',
]






# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio



import numpy as np


def shuffle3d_channels(arr):
    """
    Shuffles the channels of a 3D array and returns a flattened result.
    
    Parameters:
    - arr (numpy.ndarray): Input 3D array of shape (frames, channels, samples_per_frame)
    
    Returns:
    - numpy.ndarray: Flattened array with interleaved channels.
    """
    frames, channels, samples_per_frame = arr.shape
    # Reshape to (frames * samples_per_frame, channels)
    reshaped = arr.transpose(0, 2, 1).reshape(-1, channels)
    # Interleave channels and flatten
    return reshaped.flatten()


# @Mem.master.add
def shuffle2d_channels(arr):
    """
    Shuffles the channels of a 2D array and returns a flattened result.

    Parameters:
    - arr (numpy.ndarray): Input 2D array of shape (m, n), where m and n are dimensions.

    Returns:
    - numpy.ndarray: Flattened array with shuffled channels.
    """
    return arr.T.reshape(-1)


def get_mute_mode_data(nchannel, nperseg):
        if nchannel < 2:
            return np.zeros((nperseg), 'f')
        else:
            return np.zeros((nchannel, nperseg), 'f')
        

def map_channels(in_data:np.ndarray, in_channels, out_channels):
    """
    Map input audio channels to desired output channels.

    Args:
    in_data (np.ndarray): Input audio data.
    in_channels (int): Number of input channels.
    out_channels (int): Number of desired output channels.
    data_chunk (int): Size of data chunk for processing.

    Returns:
    np.ndarray: Processed audio data with desired number of channels.
    """

    if in_channels == 1:
        output = np.expand_dims(in_data, 0)
    else:
        # Reshape multi-channel data
        output = in_data.reshape(-1, in_channels).T

    # Upmixing
    if in_channels < out_channels:
        # Duplicate last channel for additional output channels
        output = np.vstack((output, np.tile(output[-1], (out_channels - in_channels, 1))))

    # Downmixing
    elif in_channels > out_channels:
        output = np.mean(output, axis=0, keepdims=True)
        output = np.tile(output[-1], (in_channels - out_channels, 1))

        # if in_channels == 2 and out_channels == 1:
        #     # Stereo to mono
        #     output = np.mean(output, axis=0, keepdims=True)
        # else:
        #     # General downmixing (average channels)
        #     output = output[:out_channels]
    
    return output


# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True

# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


import numpy as np
cimport numpy as np
cimport cython
from sudio.process.fx._tempo cimport _tempo_cy
from sudio.process.fx._fade_envelope cimport prepare_envelope
from scipy.interpolate import interp1d
from sudio.rateshift import ConverterType, resample, Resampler

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray pitch_shifter_cy(
        np.ndarray input_audio, 
        np.ndarray[double, ndim=1] envelope,
        float ratio=1.0, 
        int sample_rate=44100,
        bint enable_spline=True,
        float spline_sigma = 0.1,
        float fade_max_db = 0.0,
        float fade_min_db = -60.0,
        int envbuffer = 400,
        int envlen = 600,
        int frame_length = 40, 
        int sequence_ms = 82,
        int seekwindow_ms = 28, 
        int overlap_ms = 12,
        object converter_type=ConverterType.sinc_fastest
):

    if len(envelope) > 1:
        envelope = 1.0 / envelope
        envelope = prepare_envelope(
            envlen,
            envelope,
            enable_spline,
            spline_sigma,
            fade_max_db,
            fade_min_db,
            envbuffer
        )
    else:
        ratio = 1.0 / ratio
        envelope = np.full(envlen, ratio, dtype=np.float64)


    intp = interp1d(
            np.linspace(0, 1, len(envelope)), 
            envelope
        )
    
    cdef np.ndarray tempo_res = _tempo_cy(
        input_audio,
        intp,
        sample_rate=sample_rate,
        sequence_ms=sequence_ms,
        seekwindow_ms=seekwindow_ms,
        overlap_ms=overlap_ms,
    )
    
    if tempo_res.ndim == 1:
        tempo_res = tempo_res[np.newaxis, :]


    cdef np.ndarray[np.float32_t, ndim=2] result
    cdef int nchannels = tempo_res.shape[0]
    cdef int samples = tempo_res.shape[1]
    cdef int data_chunk = (sample_rate * frame_length) // 1000
    cdef int total_steps = samples // data_chunk
    cdef int current_pos
    cdef float current_ratio
    cdef np.ndarray[np.float32_t, ndim=1] frame
    cdef np.ndarray[np.float32_t, ndim=2] resampled


    if len(envelope) > 1:
        result = np.zeros((nchannels, 0), dtype=np.float32)
        resampler = Resampler(converter_type, nchannels)
        
        for i in range(total_steps):
            current_pos = i * data_chunk
            is_last_chunk = current_pos + data_chunk >= samples
            current_ratio = float(intp(float(current_pos) / samples))
            resampler.set_ratio(current_ratio)
            frame = tempo_res[:, current_pos: current_pos + data_chunk].T.flatten()
            frame = resampler.process(frame, current_ratio, is_last_chunk)
            resampled = frame.reshape(len(frame) // nchannels, nchannels).T
            result = np.concatenate((result, resampled), axis=1)
    
    else:
        result = resample(
            tempo_res,
            ratio,
            converter_type   
        )

    if nchannels == 1:
        result = result[0]
    
    return result



@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray _pitch_shifter_cy(
        np.ndarray input_audio, 
        object intp,
        int sample_rate=44100,
        int frame_length = 85, 
        int sequence_ms = 82,
        int seekwindow_ms = 28, 
        int overlap_ms = 12,
        object converter_type=ConverterType.sinc_fastest
):

    cdef np.ndarray tempo_res = _tempo_cy(
        input_audio,
        intp,
        sample_rate=sample_rate,
        sequence_ms=sequence_ms,
        seekwindow_ms=seekwindow_ms,
        overlap_ms=overlap_ms,
    )
    
    if tempo_res.ndim == 1:
        tempo_res = tempo_res[np.newaxis, :]


    cdef np.ndarray[np.float32_t, ndim=2] result
    cdef int nchannels = tempo_res.shape[0]
    cdef int samples = tempo_res.shape[1]
    cdef int data_chunk = (sample_rate * frame_length) // 1000
    cdef int total_steps = samples // data_chunk
    cdef int current_pos
    cdef float current_ratio
    cdef np.ndarray[np.float32_t, ndim=1] frame
    cdef np.ndarray[np.float32_t, ndim=2] resampled


    result = np.zeros((nchannels, 0), dtype=np.float32)
    resampler = Resampler(converter_type, nchannels)
    
    for i in range(total_steps):
        current_pos = i * data_chunk
        is_last_chunk = current_pos + data_chunk >= samples
        current_ratio = float(intp(float(current_pos) / samples))
        frame = tempo_res[:, current_pos: current_pos + data_chunk].T.flatten()
        frame = resampler.process(frame, current_ratio, is_last_chunk)
        resampled = frame.reshape(len(frame) // nchannels, nchannels).T
        result = np.concatenate((result, resampled), axis=1)
    

    if nchannels == 1:
        result = result[0]
    
    return result





# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio



import threading
import queue
import numpy as np

from sudio.types import StreamError, PipelineProcessType


class Stream:
    def __init__(
            self, 
            other, 
            obj, 
            name=None, 
            stream_type='multithreading', 
            channel=None, 
            process_type:PipelineProcessType=PipelineProcessType.MAIN
            ):

        '''
        Initialize the Stream class for managing audio data streams.

        Args:
        - other: An instance of the main audio processing class.
        - obj: An object representing the data stream (e.g., list, queue).
        - name: A name for the stream (optional).
        - stream_type: Type of streaming operation ('multithreading' or reserved for future versions).
        - channel: Channel index for controlling multi-channel streams (optional).
        - process_type: Type of stream processing PipelineProcessType 
        (PipelineProcessType.MAIN, PipelineProcessType.MULTI_STREAM, PipelineProcessType.BRANCH, PipelineProcessType.QUEUE).

        Raises:
        - StreamError: If there are issues with the provided parameters based on the selected processing mode.

        Note:
        - This class provides a flexible mechanism for managing different types of audio data streams based on the specified
          processing mode and stream type.
        '''

        self.stream_type = stream_type
        self.process_obj = other
        self.process_type = process_type
        self.pip = obj
        self.channel = channel
        self.name = name
        self._data_indexer = lambda data: data[channel]
        self._lock = threading.Lock()

        if process_type == PipelineProcessType.MAIN and type(obj) == list and len(obj) == other._nchannels and channel is None:
            self.process_type = PipelineProcessType.MULTI_STREAM

        if process_type == PipelineProcessType.MULTI_STREAM:
            if not channel is None:
                raise StreamError('In multi_stream mode, channel value must have None value (see documentation)')
            elif not type(obj) == list or not len(obj) == other._nchannels:
                raise StreamError('In multi_stream mode, pip argument must be a list with the length of the number of data channels')
            elif other._nchannels < 2:
                raise StreamError('The multi_stream mode works only when the number of input channels is set to be more than 1')

        elif process_type == PipelineProcessType.MAIN:
            if not channel is None and not (type(channel) is int and channel <= other._nchannels):
                raise StreamError('Invalid control channel data')

        if stream_type == 'multithreading':
            if process_type == PipelineProcessType.MAIN:
                self.put = self._main_put
                self.clear = obj.clear
                self.sget = self._main_get
                self.get = self._main_get

            elif process_type == PipelineProcessType.MULTI_STREAM:
                self.put = self._multi_main_put
                self.clear = self._multi_main_clear
                self.sget = self._multi_main_get
                self.get = self._multi_main_get

            elif process_type == PipelineProcessType.BRANCH:
                self.put = obj.put
                self.clear = obj.clear
                self.sget = obj.get
                self.get = obj.get

            elif process_type == PipelineProcessType.QUEUE:
                self.put = self._main_put
                self.clear = obj.queue.clear
                self.sget = self._main_get
                self.get = self._main_get

            if channel is None or (self.process_obj.nchannels == 1):
                self._data_indexer = lambda data: data
        else:
            pass


    def acquire(self):
        '''
        Acquire the lock for thread-safe operations.

        Note:
        - This method is used for thread synchronization in multithreading scenarios.
        '''
        self._lock.acquire()

    def release(self):
        '''
        Release the lock after thread-safe operations.

        Note:
        - This method is used for releasing the lock acquired during thread-safe operations.
        '''
        self._lock.release()

    def locked(self):
        '''
        Check if the lock is currently held.

        Returns:
        - bool: True if the lock is currently held, False otherwise.

        Note:
        - This method is used to check the status of the lock for thread synchronization.
        '''
        return self._lock.locked()

    def _main_put(self, data):
        '''
        Put data into the main stream and distribute it to all connected branches.

        Args:
        - data: Numpy array containing audio data.

        Notes:
        - Applies windowing to the data if specified in the main audio processing instance.
        - Distributes data to connected branches using their respective data indexers.
        '''

        data = self.process_obj._multi_channel_windowing(data)
        # print(data.shape)
        for branch in self.process_obj.branch_pipe_database:
            try:
                branch.put(branch._data_indexer(data))
            except queue.Full:
                pass
        # print(self.process_obj.branch_pipe_database)
        self.pip.put(data)

    def _main_get(self, *args, **kwargs):
        '''
        Get data from the main stream.

        Returns:
        - Numpy array containing audio data retrieved from the main stream.

        Notes:
        - Reverses windowing if specified in the main audio processing instance.
        '''
        data = self.pip.get(*args, **kwargs)
        if self.process_obj._window_type:
            data = self.process_obj._multi_channel_overlap(data)
        # print(data.shape)
        return data.astype(self.process_obj._sample_width_format_str)

    def _multi_main_put(self, data):
        '''
        Put data into the multi-stream, applying windowing if specified.

        Args:
        - data: Numpy array containing audio data.

        Notes:
        - Applies windowing to the data if specified in the main audio processing instance.
        - Distributes data to connected branches using their respective data indexers.
        '''
        # windowing
        if self.process_obj._window_type:
            final = []
            # Channel 0
            win = np.vstack((self.process_obj._windowing_buffer[0][1], np.hstack(
                (self.process_obj._windowing_buffer[0][1][self.process_obj._nhop:],
                 self.process_obj._windowing_buffer[0][0][:self.process_obj._nhop])))) * self.process_obj._window

            obj = self.process_obj._windowing_buffer[0]
            obj.pop()
            obj.insert(0, data[0])

            try:
                self.pip[0].put(win)
            except queue.Full:
                pass

            final.append(win)
            for i in range(1, self.process_obj.nchannels):
                win = np.vstack((self.process_obj._windowing_buffer[i][1], np.hstack(
                    (self.process_obj._windowing_buffer[i][1][self.process_obj._nhop:],
                     self.process_obj._windowing_buffer[i][0][:self.process_obj._nhop])))) * self.process_obj._window

                obj = self.process_obj._windowing_buffer[i]
                obj.pop()
                obj.insert(0, data[i])

                try:
                    self.pip[i].put(win)
                except queue.Full:
                    pass

                final.append(win)
        else:
            final = data.astype('float64')
            for i in range(self.process_obj.nchannels):
                try:
                    self.pip[i].put(final[i])
                except queue.Full:
                    pass

        # for 2 channel win must be an 2, 2, self.process_obj._data_chunk(e.g. 256)
        # reshaping may create some errors
        # return data
        for branch in self.process_obj.branch_pipe_database:
            branch.put(branch._data_indexer(final))

    def _multi_main_get(self, *args):
        '''
        Get data from the multi-stream.

        Returns:
        - Numpy array containing audio data retrieved from the multi-stream.

        Notes:
        - Reverses windowing if specified in the main audio processing instance.
        '''
        # channel 0
        if self.process_obj._window_type:
            win = self.pip[0].get(*args)
            retval = np.hstack(
                (self.process_obj._overlap_buffer[0][self.process_obj._nhop:], win[1][:self.process_obj._nhop])) + \
                     win[0]
            self.process_obj._overlap_buffer[0] = win[1]

            for i in range(1, self.process_obj.nchannels):
                win = self.pip[i].get(*args)
                tmp = np.hstack(
                    (self.process_obj._overlap_buffer[i][self.process_obj._nhop:], win[1][:self.process_obj._nhop])) + \
                      win[0]
                retval = np.vstack((retval, tmp))
                self.process_obj._overlap_buffer[i] = win[1]
        else:
            retval = self.pip[0].get(*args)
            for i in range(1, self.process_obj.nchannels):
                tmp = self.pip[i].get(*args)
                retval = np.vstack((retval, tmp))

        return retval.astype(self.process_obj._sample_width_format_str)

    def _multi_main_clear(self):
        '''
        Clear all connected pipes in the multi-stream.
        '''
        for i in self.pip:
            i.clear()

    def set(self, other):
        '''
        Set the attributes of the current stream based on another stream.

        Args:
        - other: Another instance of the Stream class.

        Note:
        - This method allows setting the attributes of the current stream to match those of another stream.
        '''
        if isinstance(other, self.__class__):
            assert ((self.process_type == PipelineProcessType.MAIN) or
                    (self.process_type == PipelineProcessType.MULTI_STREAM) or
                    (self.process_type == PipelineProcessType.QUEUE)) \
                   and \
                   ((other.process_type == PipelineProcessType.MAIN) or
                    (other.process_type == PipelineProcessType.MULTI_STREAM) or
                    (other.process_type == PipelineProcessType.QUEUE)), \
                'set is enabled only for "main or multi_stream" modes'
            (self.put,
             self.clear,
             self.get,
             self.sget,
             self.stream_type,
             self.process_obj,
             self.process_type,
             self.pip,
             self.channel,
             self.name) = (other.put,
                           other.clear,
                           other.get,
                           other.sget,
                           other.stream_type,
                           other.process_obj,
                           other.process_type,
                           other.pip,
                           other.channel,
                           other.name)

    def __eq__(self, other):
        '''
        Check if two Stream instances are equal.

        Returns:
        - bool: True if the two instances are equal, False otherwise.

        Note:
        - This method compares key attributes to determine equality between two Stream instances.
        '''
        if isinstance(other, self.__class__):
            return (self.put, self.clear, self.get) == (other.put, other.clear, other.get)
        return False

    def __ne__(self, other):
        '''
        Check if two Stream instances are not equal.

        Returns:
        - bool: True if the two instances are not equal, False otherwise.

        Note:
        - This method is the negation of the __eq__ method.
        '''
        return not self.__eq__(other)

    def type(self):
        '''
        Get the type of stream processing.

        Returns:
        - PipelineProcessType: The type of stream processing (MAIN, MULTI_STREAM, BRANCH, QUEUE).

        Note:
        - This method returns the type of stream processing used by the current instance.
        '''
        return self.process_type



# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


from sudio.io import SampleFormat

class FX:
    def __init__(
            self, 
            *args,
            data_size: int = None,
            sample_rate: int = None,
            nchannels: int = None,
            sample_format: SampleFormat = SampleFormat.UNKNOWN,
            data_nperseg:int=None,
            sample_type: str = '',
            sample_width: int = None,
            streaming_feature: bool = True, 
            offline_feature: bool = True, 
            preferred_datatype: SampleFormat = SampleFormat.UNKNOWN,
            **kwargs,
            ) -> None:
        
        """
        Initialize the base Effects (FX) processor with audio configuration and processing features.

        This method sets up the fundamental parameters and capabilities for audio signal processing,
        providing a flexible foundation for various audio effects and transformations.

        Parameters:
        -----------
        data_size : int, optional
            Total size of the audio data in samples. Helps in memory allocation and processing planning.

        sample_rate : int, optional
            Number of audio samples processed per second. Critical for time-based effects and analysis.

        nchannels : int, optional
            Number of audio channels (mono, stereo, etc.). Determines multi-channel processing strategies.

        sample_format : SampleFormat, optional
            Represents the audio data's numeric representation and precision. 
            Defaults to UNKNOWN if not specified.

        data_nperseg : int, optional
            Number of samples per segment, useful for segmented audio processing techniques.

        sample_type : str, optional
            Additional type information about the audio samples.

        sample_width : int, optional
            Bit depth or bytes per sample, influencing audio resolution and dynamic range.

        streaming_feature : bool, default True
            Indicates if the effect supports real-time, streaming audio processing.

        offline_feature : bool, default True
            Determines if the effect can process entire audio files or large datasets.

        preferred_datatype : SampleFormat, optional
            Suggested sample format for optimal processing. Defaults to UNKNOWN.

        Notes:
        ------
        This base class provides a standardized interface for audio effect processors,
        enabling consistent configuration and feature detection across different effects.
        """
        self._streaming_feature = streaming_feature
        self._offline_feature = offline_feature
        self._preferred_datatype = preferred_datatype
        self._data_size = data_size
        self._sample_rate = sample_rate
        self._nchannels = nchannels
        self._sample_format = sample_format
        self._data_nperseg = data_nperseg
        self._sample_type = sample_type
        self._sample_width = sample_width

    def is_streaming_supported(self) -> bool:
        """
        Determine if audio streaming is supported for this effect.
        """
        return self._streaming_feature

    def is_offline_supported(self) -> bool:
        """
        Check if file/batch audio processing is supported.
        """
        return self._offline_feature

    def get_preferred_datatype(self)->SampleFormat:
        """
        Retrieve the recommended sample format for optimal processing.
        """
        return self._preferred_datatype
    
    
    def get_data_size(self)->int:
        """
        Get the total size of audio data in samples.
        """
        return self._data_size
    
    def get_sample_rate(self)->int:
        """
        Retrieve the audio sampling rate.
        """
        return self._sample_rate
    
    def get_nchannels(self)->int:
        """
        Get the number of audio channels.
        """
        return self._nchannels
    
    def get_sample_format(self)->SampleFormat:
        """
        Retrieve the audio sample format.
        """
        return self._sample_format
    
    def get_sample_type(self)->str:
        """
        Get additional sample type information.
        """
        return self._sample_type
    
    def get_sample_width(self):
        """
        Retrieve the bit depth or bytes per sample.
        """
        return self._sample_width

    def process(*args, **kwargs):
        """
        Base method for audio signal processing.
        
        This method should be implemented by specific effect classes
        to define their unique audio transformation logic.
        """
        ...

        



# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio



import numpy as np
from sudio.rateshift import ConverterType, resample
from sudio.io import SampleFormat
from sudio.io import get_sample_size
from sudio.utils.channel import shuffle2d_channels
from sudio.metadata import AudioMetadata


def synchronize_audio(rec: AudioMetadata,
                      nchannels: int,
                      sample_rate: int,
                      sample_format_id: int,
                      output_data='byte') -> dict:
    """
    Synchronizes and transforms audio recording parameters.

    Modifies audio recording by adjusting channels, resampling, 
    and converting sample format.

    Args:
    -----

        - rec: Audio recording metadata
        - nchannels: Desired number of audio channels
        - sample_rate: Target sample rate
        - sample_format_id: Desired audio sample format
        - output_data: Output data format ('byte' or 'ndarray')

    Returns:
    --------

        - Modified audio recording metadata
    """

    form = get_sample_size(rec.sampleFormat)
    if rec.sampleFormat == SampleFormat.FLOAT32:
        form = '<f{}'.format(form)
    else:
        form = '<i{}'.format(form)
    data = np.frombuffer(rec.o, form)
    if rec.nchannels == 1:
        if nchannels > rec.nchannels:
            data = np.vstack([data for i in range(nchannels)])
            rec.nchannels = nchannels

    else:
        # Safety update: Ensure all arrays have the same size
        channel_data = [data[i::rec.nchannels] for i in range(nchannels)]
        min_length = min(len(channel) for channel in channel_data)
        channel_data = [channel[:min_length] for channel in channel_data]
        data = np.array(channel_data)

    if not sample_rate == rec.frameRate:
        scale = sample_rate / rec.frameRate
        dtype = data.dtype
        data = data.astype(np.float32)
        data = resample(data, scale, ConverterType.sinc_fastest)
        data.astype(dtype)

    if output_data.startswith('b') and rec.nchannels > 1:
        data = shuffle2d_channels(data)

    rec.nchannels = nchannels
    rec.sampleFormat = sample_format_id

    form = get_sample_size(sample_format_id)
    if sample_format_id == SampleFormat.FLOAT32:
        form = '<f{}'.format(form)
    else:
        form = '<i{}'.format(form)

    if output_data.startswith('b'):
        rec.o = data.astype(form).tobytes()
    else:
        rec.o = data.astype(form)

    rec.size = len(rec.o)
    rec.frameRate = sample_rate

    rec.duration = rec.size / (rec.frameRate *
                                     rec.nchannels *
                                     get_sample_size(rec.sampleFormat))

    return rec




# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio



class AudioRecordDatabase:
    """
    A lightweight in-memory database for managing audio records.

    Provides dictionary-like access and manipulation of audio records
    with simple key (name) based storage and retrieval.
    """

    def __init__(self):
        """
        Initialize an empty audio record database.

        Creates an empty dictionary to store audio records with names as keys.
        """
        self.records = {}

    def add_record(self, record):
        """
        Add an audio record to the database.

        Args:
        -----

            record: An audio record with a 'name' attribute to be used as the key.
        """
        self.records[record.name] = record

    def get_record(self, name):
        """
        Retrieve a specific audio record by its name.

        Args:
        -----
            
            name (str): Name of the audio record to retrieve.

        Returns:
        --------

            The audio record if found, None otherwise.
        """

        return self.records.get(name)

    def remove_record(self, name):
        """
        Remove a record from the database by its name.

        Args:
        -----

            name (str): Name of the audio record to remove.
        """
        if name in self.records:
            del self.records[name]

    def index(self):
        """
        Get a list of all record names in the database.

        Returns:
        --------

            list: Names of all records currently in the database.
        """
        return list(self.records.keys())

    def __getitem__(self, name):
        """
        Allow dictionary-style access to records.

        Args:
        -----

            name (str): Name of the record to retrieve.

        Returns:
        --------

            The audio record corresponding to the name.
        """
        return self.get_record(name)

    def __setitem__(self, name, record):
        """
        Allow dictionary-style record addition.

        Args:
        -----

            - name (str): Name to associate with the record.
            - record: The audio record to store.
        """
        self.add_record(record)



# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


from sudio.utils.strtool import generate_timestamp_name


class TimedIndexedString:
    def __init__(self,
                 string: str,
                 start_from: int = None,
                 first_index: int = 0,
                 start_before: str = None,
                 max_index: int = None,
                 seed: str = ''):
        """
        A class for generating indexed strings with optional time-based suffixes.

        Args:
            base_string (str): The initial string.
            start_from (int, optional): Starting index. Defaults to None.
            first_index (int, optional): Initial index. Defaults to 0.
            start_before (str, optional): Character before which the index should start. Defaults to None.
            max_index (int, optional): Maximum index. Defaults to None.
            seed (str, optional): Seed for additional uniqueness. Defaults to ''.
        """
        if start_from is not None:
            index = start_from
        elif start_before is not None:
            index = string.index(start_before)
        else:
            raise ValueError("Either 'start_from' or 'start_before' must be provided.")

        self._count = first_index
        self._first_index = first_index
        self._max_index = max_index
        self._str: list = [(string[:index] + '_' + seed + '_') if seed and (seed not in string) else string[:index] + '_',
                           '',
                           string[index:]]
        self._current_str = ''

    def __call__(self, *args, timed_random: bool = False, timed_regular: bool = False, seed: str = ''):
        """
        Generates an indexed string.

        Args:
            *args: Additional string elements to append to the index.
            timed_random (bool, optional): Whether to add a timed random suffix. Defaults to False.
            timed_regular (bool, optional): Whether to add a timed regular suffix. Defaults to False.
            seed (str, optional): Seed for additional uniqueness. Defaults to ''.

        Returns:
            str: The generated indexed string.

        Raises:
            OverflowError: If the maximum index is reached.

        Example:
            >>> indexed_name = TimedIndexedString("example_", start_before="_", max_index=3)
            >>> indexed_name("suffix")  # 'example_0_suffix'
            >>> indexed_name("extra", timed_regular=True)  # 'example_1_extra_YYYYMMDD_SSMMM'
        """
        if self._max_index and self._count > self._max_index:
            raise OverflowError("Maximum index reached.")

        self._str[1] = str(self._count) + ''.join(args)
        self._count += 1

        if seed:
            self._str[0] += '_' + seed

        if timed_random:
            self._str[1] += generate_timestamp_name('_')
        elif timed_regular:
            self._str[1] += '_' + generate_timestamp_name()

        self._current_str = ''.join(self._str)
        return self._current_str

    def reset(self):
        """
        Resets the index counter to its initial value.
        """
        self._count = self._first_index

    def get(self):
        """
        Gets the current indexed string.

        Returns:
            str: The current indexed string.
        """
        return self._current_str



# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


from sudio.process.fx import FX
from sudio.io import SampleFormat
import numpy as np
from typing import TYPE_CHECKING
from  sudio.process.fx._tempo import tempo_cy

class Tempo(FX):
    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize the Tempo audio effect processor for time stretching.

        Configures time stretching with support for both streaming and offline 
        audio processing, optimized for 32-bit floating-point precision.

        Notes:
        ------
        Implements advanced time stretching using WSOLA (Waveform Similarity 
        Overlap-Add) algorithm to modify audio tempo without altering pitch.
        """
        features = {
            'streaming_feature': True, 
            'offline_feature': True,
            'preferred_datatype': SampleFormat.FLOAT32
            }
        super().__init__(*args, **kwargs, **features)

    def process(self, data: np.ndarray, tempo:float=1.0, envelope:np.ndarray=[], **kwargs):
        """
        Perform time stretching on the input audio data without altering pitch.

        This method allows tempo modification through uniform or dynamic tempo changes,
        utilizing an advanced Waveform Similarity Overlap-Add (WSOLA) algorithm to 
        manipulate audio duration while preserving sound quality and spectral characteristics.

        Parameters:
        -----------
        data : np.ndarray
            Input audio data as a NumPy array. Supports mono and multi-channel audio.
            Recommended data type is float32.

        tempo : float, optional
            Tempo scaling factor for time stretching.
            - 1.0 means no change in tempo/duration
            - < 1.0 slows down audio (increases duration)
            - > 1.0 speeds up audio (decreases duration)
            Default is 1.0.

            Examples:
            - 0.5: doubles audio duration
            - 2.0: halves audio duration

        envelope : np.ndarray, optional
            Dynamic tempo envelope for time-varying tempo modifications.
            Allows non-uniform tempo changes across the audio signal.
            Default is an empty list (uniform tempo modification).

            Example:
            - A varying array of tempo ratios can create complex time-stretching effects

        **kwargs : dict
            Additional keyword arguments passed to the underlying tempo algorithm.
            Allows fine-tuning of advanced parameters such as:
            - sequence_ms: Sequence length for time-stretching window
            - seekwindow_ms: Search window for finding similar waveforms
            - overlap_ms: Crossfade overlap between segments
            - enable_spline: Enable spline interpolation for envelope
            - spline_sigma: Gaussian smoothing parameter for envelope

        Returns:
        --------
        np.ndarray
            Time-stretched audio data with the same number of channels and original data type
            as the input.

        Examples:
        ---------
        >>> slow_audio = tempo_processor.process(audio_data, tempo=0.5)  # Slow down audio
        >>> fast_audio = tempo_processor.process(audio_data, tempo=1.5)  # Speed up audio
        >>> dynamic_tempo = tempo_processor.process(audio_data, envelope=[0.5, 1.0, 2.0])  # Dynamic tempo

        Notes:
        ------
        - Preserves audio quality with minimal artifacts
        - Uses advanced WSOLA algorithm for smooth time stretching
        - Supports both uniform and dynamic tempo modifications
        - Computationally efficient implementation
        - Does not change the pitch of the audio

        Warnings:
        ---------
        - Extreme tempo modifications (very low or high values) may introduce 
        audible artifacts or sound distortions
        - Performance and quality may vary depending on audio complexity
        """

        dtype = data.dtype
        data = tempo_cy(
            data, 
            np.asarray(envelope, dtype=np.double), 
            self._sample_rate, 
            default_tempo=tempo, 
            **kwargs
            )
        return data.astype(dtype)
    
    


# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True


# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio



import numpy as np
cimport numpy as np
cimport cython

DTYPE = np.float32
ctypedef np.float32_t DTYPE_t

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)

def channel_mixer(
    np.ndarray[DTYPE_t, ndim=2] data, 
    np.ndarray[DTYPE_t, ndim=2] correlation,
):

    cdef int nchannels = data.shape[0]
    cdef int nsamples = data.shape[1]
    
    cdef np.ndarray[DTYPE_t, ndim=2] result = np.zeros((nchannels, nsamples), dtype=DTYPE)
    
    cdef int i, j, k
    cdef DTYPE_t temp
    
    for i in range(nchannels):
        for j in range(nsamples):
            temp = 0.0
            for k in range(nchannels):
                temp += correlation[i, k] * data[k, j]
            result[i, j] = temp
    
    return result





# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


import io
import os
import time
from builtins import ValueError
import numpy as np
from typing import Union,Tuple
import platform

from sudio.audiosys.sync import synchronize_audio
from sudio.utils.timed_indexed_string import TimedIndexedString
from sudio.types import DecodeError
from sudio.metadata import AudioMetadata
from sudio.io import SampleFormat, get_sample_size

def is_file_locked(file_path):
    """
    Check if a file is currently locked or in use by another process.

    Supports both Windows and Unix-like systems. On Windows, attempts to open 
    the file exclusively. On Unix systems, uses file locking mechanisms.

    Args:
    -----

        - file_path (str): Path to the file to check for lock status.

    Returns:
    --------
    
        - bool: True if the file is locked, False otherwise.
    """
    is_windows = platform.system() == "Windows"

    if not is_windows:
        try:
            import fcntl
        except ImportError:
            fcntl = None

    if is_windows:
        try:
            with open(file_path, 'r+') as f:
                return False
        except IOError:
            return True
    else:
        if fcntl:
            try:
                with open(file_path, 'rb') as file:
                    fcntl.flock(file.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
                    fcntl.flock(file.fileno(), fcntl.LOCK_UN)
                return False
            except IOError:
                return True
        else:
            try:
                with open(file_path, 'r+') as f:
                    return False
            except IOError:
                return True


def write_to_cached_file(*head,
                         head_dtype: str = 'u8',
                         data: Union[bytes, io.BufferedRandom] = None,
                         file_name: str = None,
                         file_mode: str = 'wb+',
                         buffered_random: io.BufferedRandom = None,
                         pre_truncate: bool = False,
                         pre_seek: Tuple[int, int] = None,
                         after_seek: Tuple[int, int] = None,
                         pre_flush: bool = False,
                         after_flush: bool = False,
                         size_on_output: bool = False,
                         data_chunk: int = int(1e6)) -> Union[io.BufferedRandom, Tuple[io.BufferedRandom, int]]:
    """
    Flexibly write data to a file with granular control over the writing process.

    Provides advanced file writing capabilities, allowing precise manipulation 
    of file operations before, during, and after writing data.

    Args:
    -----

        - *head: Variable header arguments to write at the beginning of the file.
        - head_dtype (str, optional): Numpy data type for header arguments. Defaults to 'u8' (64-bit unsigned integer).
        - data (bytes or BufferedRandom, optional): Data to write to the file. Can be bytes or a file-like object.
        - file_name (str, optional): Name of the file to write. Required if no buffered_random is provided.
        - file_mode (str, optional): Mode to open the file. Defaults to 'wb+' (write binary, read/write).
        - buffered_random (BufferedRandom, optional): Existing file object to write to. Overrides file_name if provided.
        - pre_truncate (bool, optional): Truncate the file before writing. Defaults to False.
        - pre_seek (tuple, optional): Seek to this position before writing. Tuple of (offset, whence).
        - after_seek (tuple, optional): Seek to this position after writing. Tuple of (offset, whence).
        - pre_flush (bool, optional): Flush the file before writing. Defaults to False.
        - after_flush (bool, optional): Flush the file after writing. Defaults to False.
        - size_on_output (bool, optional): Return total bytes written along with file object. Defaults to False.
        - data_chunk (int, optional): Size of chunks when reading BufferedRandom data. Defaults to 1,000,000 bytes.

    Returns:
    --------

        io.BufferedRandom or tuple: 
        - File object if size_on_output is False
        - Tuple of (file object, total bytes written) if size_on_output is True
    
    Raises:
    -------

        ValueError: If neither buffered_random nor file_name is provided.
    """
    if buffered_random:
        file: io.BufferedRandom = buffered_random
    elif file_name:
        file = open(file_name, file_mode)
    else:
        raise ValueError("Either buffered_random or file_name must be provided.")
    size = 0

    if pre_seek:
        file.seek(*pre_seek)

    if pre_truncate:
        file.truncate()

    if pre_flush:
        file.flush()

    if head:
        size = file.write(np.asarray(head, dtype=head_dtype))

    if data:
        if isinstance(data, io.BufferedRandom):
            data_position = data.tell()
            buffer = data.read(data_chunk)
            while buffer:
                size += file.write(buffer)
                buffer = data.read(data_chunk)
            data.seek(data_position, 0)
        else:
            size += file.write(data)

    if after_seek:
        file.seek(*after_seek)

    if after_flush:
        file.flush()

    if size_on_output:
        return file, size
    return file


def handle_cached_record(record: Union[AudioMetadata, dict],
                         path_server: TimedIndexedString,
                         orphaned_cache:list,
                         cache_info_size:str,
                         buffer_type:str=None,
                         decoder: callable = None,
                         sync_sample_format_id: int = 1,
                         sync_nchannels: int = 2,
                         sync_sample_rate: int = 44100,
                         safe_load: bool = True,
                         max_attempts=5) -> AudioMetadata:
    """
   Manage audio record caching with advanced synchronization and error recovery.

   Handles complex file caching scenarios for audio records, including 
   format conversion, file locking management, and safe loading strategies.

   Args:
   -----

       - record (AudioMetadata or dict): Audio record to be processed and cached.
       - path_server (TimedIndexedString): Service for generating unique file paths.
       - orphaned_cache (list): List tracking obsolete cache entries.
       - cache_info_size (str): Size of metadata in the cache.
       - buffer_type (str, optional): Type of buffer for filename parsing.
       - decoder (callable, optional): Custom function to decode audio if needed.
       - sync_sample_format_id (int, optional): Target audio sample format. Defaults to 1.
       - sync_nchannels (int, optional): Target number of audio channels. Defaults to 2.
       - sync_sample_rate (int, optional): Target audio sample rate. Defaults to 44100.
       - safe_load (bool, optional): Enable safe loading with error recovery. Defaults to True.
       - max_attempts (int, optional): Maximum retry attempts for file operations. Defaults to 5.

   Returns:
   --------

       - AudioMetadata: Processed audio record with updated metadata and file reference.

   Raises:
   -------

       IOError: If file cannot be accessed after max attempts.
       DecodeError: If audio decoding fails and no decoder is provided.
   """

    path: str = path_server()
    if path in orphaned_cache:
        attempt = 0
        while attempt < max_attempts:
            try:
                if not is_file_locked(path):
                    f = open(path, 'rb+')
                    record.o = f
                    break
                else:
                    new_path = path_server()
                    if not os.path.exists(new_path) or not is_file_locked(new_path):
                        # Write a new file based on the new path
                        data_chunk = int(1e7)
                        with open(path, 'rb') as pre_file:
                            pre_file.seek(0, 0)
                            f = open(new_path, 'wb+')
                            data = pre_file.read(data_chunk)
                            while data:
                                f.write(data)
                                data = pre_file.read(data_chunk)
                            record.o = f
                        break
            except (IOError, OSError) as e:
                attempt += 1
                time.sleep(0.1)  # Short delay before retrying

        if attempt >= max_attempts:
            raise IOError(f"Unable to access the file after {max_attempts} attempts")

        f.seek(0, 0)
        orphaned_cache = f.read(cache_info_size)
        try:
            csize, csample_rate, csample_format_id, cnchannels = np.frombuffer(orphaned_cache, dtype='u8').tolist()
            csample_format = SampleFormat(csample_format_id) if csample_format_id else SampleFormat.UNKNOWN
            
        except ValueError:
            # bad orphaned_cache error
            f.close()
            os.remove(f.name)
            raise DecodeError

        csample_format = csample_format if csample_format else None
        record.size = csize

        record.frameRate = csample_rate
        record.nchannels = cnchannels
        record.sampleFormat = csample_format

        if (cnchannels == sync_nchannels and
                csample_format == sync_sample_format_id and
                csample_rate == sync_sample_rate):
            pass

        elif safe_load:
            record.o = f.read()
            record = synchronize_audio(record,
                                        sync_nchannels,
                                        sync_sample_rate,
                                        sync_sample_format_id)

            record.o = write_to_cached_file(record.size,
                    record.frameRate,
                    record.sampleFormat if record.sampleFormat else 0,
                    record.nchannels,
                    buffered_random=f,
                    data=record.o,
                    pre_seek=(0, 0),
                    pre_truncate=True,
                    pre_flush=True,
                    after_seek=(cache_info_size, 0),
                    after_flush=True
                    )


    else:
         # Handle decoding error
        if decoder is not None:
            record.o = decoder()
        if isinstance(record.o, io.BufferedRandom):
            record.size = os.path.getsize(record.o.name)
        else:
            record.size = len(record.o) + cache_info_size

        record.o = write_to_cached_file(
            record.size,
            record.frameRate,
            record.sampleFormat if record.sampleFormat else 0,
            record.nchannels,
            file_name=path,
            data=record.o,
            pre_seek=(0, 0),
            pre_truncate=True,
            pre_flush=True,
            after_seek=(cache_info_size, 0),
            after_flush=True
            )

    record.duration = record.size / (record.frameRate *
                                           record.nchannels *
                                           get_sample_size(record.sampleFormat))

    if buffer_type is not None:
        post = record.o.name.index(buffer_type)
        pre = max(record.o.name.rfind('\\'), record.o.name.rfind('/'))
        pre = (pre + 1) if pre > 0 else 0
        record.name = record.o.name[pre: post]

    return record




cimport numpy as cnp


# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio




cpdef int find_nearest_divisible(int reference_number, int divisor)
cpdef int find_nearest_divisor(int num, int divisor) except? -1
cpdef db2amp(db)
cpdef amp2db(amp)




# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


class StreamingTimeController:
    def __get__(self, instance, owner):
        assert instance.isready(), PermissionError('current object is not streaming')
        return instance._itime_calculator(instance._stream_file.tell())

    def __set__(self, instance, tim):
        assert abs(tim) < instance.duration, OverflowError('input time must be less than the record duration')
        assert instance.isready(), PermissionError('current object is not streaming')
        seek = instance._time_calculator(abs(tim))
        if tim < 0:
            seek = instance._stream_file_size - seek
        instance._stream_file.seek(seek, 0)



# cython: language_level=3

# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio

cimport numpy as np


# cdef int DEFAULT_SEQUENCE_MS = 82
# cdef int DEFAULT_SEEKWINDOW_MS = 28
# cdef int DEFAULT_OVERLAP_MS = 12


cpdef np.ndarray tempo_cy(
    np.ndarray input_audio,
    np.ndarray[double, ndim=1] envelope,
    int sample_rate=*,
    int sequence_ms=*,
    int seekwindow_ms=*,
    int overlap_ms=*,
    bint enable_spline=*,
    double spline_sigma=*,
    double fade_max_db=*,
    double fade_min_db=*,
    int envbuffer=*,
    int envlen=*,
    double default_tempo=*
)


cpdef np.ndarray _tempo_cy(
    np.ndarray input_audio,
    object intp,
    int sample_rate=*,
    int sequence_ms=*,
    int seekwindow_ms=*, 
    int overlap_ms=*,
)




# distutils: language=c++
# cython: language_level=3


# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio

# distutils: language=c++
# cython: language_level=3

import numpy as np
cimport numpy as np
cimport cython
cimport sudio.process.fx._fade_envelope as fade_envelope

from libc.math cimport sqrt
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter1d
from sudio.utils.math cimport db2amp

DEF DEFAULT_SEQUENCE_MS = 82
DEF DEFAULT_SEEKWINDOW_MS = 28  
DEF DEFAULT_OVERLAP_MS = 12


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray tempo_cy(
    np.ndarray input_audio,
    np.ndarray[double, ndim=1] envelope,
    int sample_rate=44100,
    int sequence_ms=DEFAULT_SEQUENCE_MS,
    int seekwindow_ms=DEFAULT_SEEKWINDOW_MS, 
    int overlap_ms=DEFAULT_OVERLAP_MS,
    bint enable_spline = False,
    double spline_sigma = 0.1,
    double fade_max_db = 0.0,
    double fade_min_db = -60.0,
    int envbuffer = 400,
    int envlen = 600,
    double default_tempo = 1.0
):

    
    input_audio = np.asarray(input_audio, dtype=np.float32)
    if input_audio.ndim == 1:
        input_audio = input_audio[np.newaxis, :]
    
    cdef:
        np.ndarray[double, ndim=1] tempo_values
        
    
    if len(envelope) > 1:
       tempo_values =  fade_envelope.prepare_envelope(
            envlen,
            envelope,
            enable_spline,
            spline_sigma,
            fade_max_db,
            fade_min_db,
            envbuffer
        )
    else:
        tempo_values = np.full(envlen, default_tempo, dtype=np.float64)

    intp = interp1d(
            np.linspace(0, 1, len(tempo_values)), 
            tempo_values
        )

    cdef np.ndarray result =  _tempo_cy(
            input_audio,
            intp,
            sample_rate,
            sequence_ms,
            seekwindow_ms,
            overlap_ms
        )
    
    return result




@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray _tempo_cy(
    np.ndarray input_audio,
    object intp,
    int sample_rate=44100,
    int sequence_ms=DEFAULT_SEQUENCE_MS,
    int seekwindow_ms=DEFAULT_SEEKWINDOW_MS, 
    int overlap_ms=DEFAULT_OVERLAP_MS,
):

    
    input_audio = np.asarray(input_audio, dtype=np.float32)
    if input_audio.ndim == 1:
        input_audio = input_audio[np.newaxis, :]
    
    cdef:
        int channels = input_audio.shape[0]
        int frames = input_audio.shape[1]
        int overlap_length = (sample_rate * overlap_ms) // 1000
        int sequence_length = (sample_rate * sequence_ms) // 1000
        int seekwindow_length = (sample_rate * seekwindow_ms) // 1000
        int output_frames = frames * 2  # Generous initial allocation
        int input_pos = 0
        int output_pos = 0
        int best_offset
        double corr, best_corr
        int i, ch
        float scale1, scale2
        float[:] signal1_view
        float[:] signal2_view
        np.ndarray[np.float32_t, ndim=2] output_buffer = np.zeros((channels, output_frames), dtype=np.float32)
        np.ndarray[np.float32_t, ndim=2] mid_buffer = np.zeros((channels, overlap_length), dtype=np.float32)
        
        double current_tempo
        double skip_fract = 0.0
        int skip
        double nominal_skip
        double input_progress_ratio
    

    while input_pos + seekwindow_length < frames:
        input_progress_ratio = float(input_pos) / frames
        current_tempo = float(intp(input_progress_ratio))
        
        nominal_skip = current_tempo * (sequence_length - overlap_length)
        skip = int(skip_fract + nominal_skip + 0.5)
        skip_fract += nominal_skip - skip
        
        best_offset = 0
        best_corr = -1.0
        for i in range(seekwindow_length - overlap_length):
            signal1_view = input_audio[0, input_pos + i:input_pos + i + overlap_length]
            signal2_view = mid_buffer[0, :overlap_length]
            corr = calc_correlation(signal1_view, signal2_view)
            if corr > best_corr:
                best_corr = corr
                best_offset = i
        
        for ch in range(channels):
            for i in range(overlap_length):
                scale1 = float(i) / overlap_length
                scale2 = 1.0 - scale1
                output_buffer[ch, output_pos + i] = (
                    input_audio[ch, input_pos + best_offset + i] * scale1 + 
                    mid_buffer[ch, i] * scale2
                )
        
        sequence_offset = input_pos + best_offset + overlap_length
        sequence_length_current = min(sequence_length - overlap_length, 
                                   frames - sequence_offset)
        
        if sequence_length_current > 0:
            for ch in range(channels):
                output_buffer[ch, output_pos + overlap_length:
                            output_pos + overlap_length + sequence_length_current] = \
                    input_audio[ch, sequence_offset:
                              sequence_offset + sequence_length_current]
        
        if sequence_offset + sequence_length_current - overlap_length < frames:
            for ch in range(channels):
                mid_buffer[ch, :] = input_audio[ch,
                    sequence_offset + sequence_length_current - overlap_length:
                    sequence_offset + sequence_length_current]
        
        input_pos += skip
        output_pos += sequence_length_current
    
    #trim output buffer  mono/multi-channel
    result = output_buffer[:, :output_pos]
    if input_audio.shape[0] == 1:
        result = result[0]
        
    return result


@cython.boundscheck(False)
@cython.wraparound(False)
cdef double calc_correlation(float[:] signal1, float[:] signal2) nogil:
    """Calculate normalized cross-correlation between two signals"""
    cdef:
        int length = signal1.shape[0]
        int i
        double corr = 0.0
        double norm1 = 0.0
        double norm2 = 0.0
        
    for i in range(length):
        corr += signal1[i] * signal2[i]
        norm1 += signal1[i] * signal1[i]
        norm2 += signal2[i] * signal2[i]
        
    if norm1 == 0.0 or norm2 == 0.0:
        return 0.0
        
    return corr / sqrt(norm1 * norm2)



# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio

from sudio.process.fx import FX
from  sudio.process.fx._pitch_shifter import pitch_shifter_cy
from sudio.io import SampleFormat
import numpy as np

class PitchShifter(FX):
    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize the PitchShifter audio effect processor.

        This method configures the PitchShifter effect with specific processing features,
        setting up support for both streaming and offline audio processing.
        """
        
        features = {
            'streaming_feature': True, 
            'offline_feature': True,
            'preferred_datatype': SampleFormat.FLOAT32
            }
        super().__init__(*args, **kwargs, **features)

    def process(
            self, 
            data: np.ndarray, 
            semitones:np.float32=0.0, 
            cent:np.float32=0.0, 
            ratio:np.float32=1.0, 
            envelope:np.ndarray=[],
            **kwargs
            ):

        """
        Perform pitch shifting on the input audio data.

        This method allows pitch modification through multiple parametrization approaches:
        1. Semitone and cent-based pitch shifting
        2. Direct ratio-based pitch shifting
        3. Envelope-based dynamic pitch shifting

        Parameters:
        -----------
        data : np.ndarray
            Input audio data as a NumPy array. Supports mono and multi-channel audio.
            Recommended data type is float32.

        semitones : np.float32, optional
            Number of semitones to shift the pitch. 
            Positive values increase pitch, negative values decrease pitch.
            Default is 0.0 (no change).
            
            Example:
            - 12.0 shifts up one octave
            - -12.0 shifts down one octave

        cent : np.float32, optional
            Fine-tuning pitch adjustment in cents (1/100th of a semitone).
            Allows precise micro-tuning between semitones.
            Default is 0.0.

            Example:
            - 50.0 shifts up half a semitone
            - -25.0 shifts down a quarter semitone

        ratio : np.float32, optional
            Direct pitch ratio modifier. 
            - 1.0 means no change
            - > 1.0 increases pitch
            - < 1.0 decreases pitch
            Default is 1.0.

            Note: When semitones or cents are used, this ratio is multiplicative.

        envelope : np.ndarray, optional
            Dynamic pitch envelope for time-varying pitch shifting.
            If provided, allows non-uniform pitch modifications across the audio.
            Default is an empty list (uniform pitch shifting).

            Example:
            - A varying array of ratios can create complex pitch modulations

        **kwargs : dict
            Additional keyword arguments passed to the underlying pitch shifting algorithm.
            Allows fine-tuning of advanced parameters like:
            - sample_rate: Audio sample rate
            - frame_length: Processing frame size
            - converter_type: Resampling algorithm

        Returns:
        --------
        np.ndarray
            Pitch-shifted audio data with the same number of channels as input.

        Examples:
        ---------
        >>> record = record.afx(PitchShifter, start=30, envelope=[1, 3, 1, 1]) # Dynamic pitch shift
        >>> record = record.afx(PitchShifter, semitones=4)  # Shift up 4 semitones
        
        Notes:
        ------
        - Uses high-quality time-domain pitch shifting algorithm
        - Preserves audio quality with minimal artifacts
        - Supports both uniform and dynamic pitch modifications
        """
       
        # pitch ratio based on semitones and cents
        if semitones != 0.0 or cent != 0.0:
            pitch_ratio = pow(2.0, (semitones + cent / 100) / 12.0) * ratio
        else:
            pitch_ratio =  ratio

        res = pitch_shifter_cy(
            data,
            np.asarray(envelope, dtype=np.double),
            sample_rate=self._sample_rate,
            ratio=pitch_ratio,
            **kwargs
        )

        return res
    
    
    


# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio

from typing import Union, List, Optional
import numpy as np
from sudio.process.fx import FX
from sudio.io import SampleFormat
from sudio.process.fx._channel_mixer import channel_mixer


class ChannelMixer(FX):
    def __init__(
            self, *args, 
            **kwargs) -> None:
        """
        initialize the ChannelMixer audio effect processor.

        Parameters:
        -----------
        *args : Variable positional arguments
            Arguments to be passed to the parent FX class initializer.
        **kwargs : dict, optional
            Additional keyword arguments for configuration.

        """
        features = {
            'streaming_feature': True, 
            'offline_feature': True,
            'preferred_datatype': SampleFormat.FLOAT32
        }
        super().__init__(*args, **kwargs, **features)

    def process(
        self, 
        data: np.ndarray, 
        correlation: Optional[Union[List[List[float]], np.ndarray]] = None,
        **kwargs
    ) -> np.ndarray:
        """
        apply channel mixing to the input audio signal based on a correlation matrix.

        manipulates multi-channel audio by applying inter-channel correlation 
        transformations while preserving signal characteristics.

        Parameters:
        -----------
        data : numpy.ndarray
            Input multi-channel audio data. Must have at least 2 dimensions.
            Shape expected to be (num_channels, num_samples).

        correlation : Union[List[List[float]], numpy.ndarray], optional
            Correlation matrix defining inter-channel relationships.
            - If None, returns input data unchanged
            - Must be a square matrix matching number of input channels
            - Values must be between -1 and 1
            - Matrix shape: (num_channels, num_channels)

        **kwargs : dict, optional
            Additional processing parameters (currently unused).

        Returns:
        --------
        numpy.ndarray
            Channel-mixed audio data with the same shape as input.

        Raises:
        -------
        ValueError
            - If input data has fewer than 2 channels
            - If correlation matrix is incorrectly shaped
            - If correlation matrix contains values outside [-1, 1]

        Examples:
        ---------
            >>> from sudio.process.fx import ChannelMixer
            >>> su = sudio.Master()        
            >>> rec = su.add('file.mp3')
            >>> newrec = rec.afx(ChannelMixer, correlation=[[.4,-.6], [0,1]]) #for two channel
        """

        if data.ndim < 2:
            raise ValueError("Input data must be multichannel")
        
        nchannels = self._nchannels

        if correlation is None:
            return data
        else:
            correlation = np.asarray(correlation, dtype=np.float32)

        if (correlation.shape != (nchannels, nchannels) or 
            not np.all((-1 <= correlation) & (correlation <= 1))):
            raise ValueError(f"Invalid correlation matrix: shape {correlation.shape}, expected {(nchannels, nchannels)}")

        return channel_mixer(data, correlation)
    





# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio



import os
import sys

def get_encoded_filename_bytes(filepath: str) -> bytes:
    """
    Encode the given file path string into bytes using the system's filesystem encoding.

    Parameters:
    - filepath (str): The input file path.

    Returns:
    - bytes: The encoded file path as bytes.

    Raises:
    - FileNotFoundError: If the specified file does not exist.
    """
    expanded_filepath = os.path.expanduser(filepath)
    
    if not os.path.isfile(expanded_filepath):
        raise FileNotFoundError(filepath)

    return expanded_filepath.encode(sys.getfilesystemencoding())



# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


import numpy as np
from sudio.process.fx import FX
from sudio.io import SampleFormat
from sudio.process.fx._fade_envelope import generate_envelope, prepare_envelope
from sudio.process.fx._fade_envelope import FadePreset as FP
from enum import Enum
from typing import Union

class FadePreset(Enum):
    SMOOTH_ENDS = FP.SMOOTH_ENDS
    BELL_CURVE = FP.BELL_CURVE
    KEEP_ATTACK_ONLY = FP.KEEP_ATTACK_ONLY
    LINEAR_FADE_IN = FP.LINEAR_FADE_IN
    LINEAR_FADE_OUT = FP.LINEAR_FADE_OUT
    PULSE = FP.PULSE
    REMOVE_ATTACK = FP.REMOVE_ATTACK
    SMOOTH_ATTACK = FP.SMOOTH_ATTACK
    SMOOTH_FADE_IN = FP.SMOOTH_FADE_IN
    SMOOTH_FADE_OUT = FP.SMOOTH_FADE_OUT
    SMOOTH_RELEASE = FP.SMOOTH_RELEASE
    TREMORS = FP.TREMORS
    ZIGZAG_CUT = FP.ZIGZAG_CUT


class FadeEnvelope(FX):
    def __init__(self, *args, **kwargs) -> None:
        features = {
            'streaming_feature': True, 
            'offline_feature': True,
            'preferred_datatype': SampleFormat.FLOAT32
        }
        """
        Initialize the FadeEnvelope audio effect processor.

        This method configures the FadeEnvelope effect with specific processing features,
        setting up support for both streaming and offline audio processing.

        Parameters:
        -----------
        *args : Variable positional arguments
            Allows passing additional arguments to the parent FX class.

        **kwargs : Variable keyword arguments
            Additional configuration parameters for the effect.

        Features:
        ---------
        - Supports streaming audio processing
        - Supports offline audio processing
        - Prefers 32-bit floating-point audio format for high-precision dynamics manipulation

        Notes:
        ------
        The FadeEnvelope effect provides versatile amplitude shaping capabilities,
        enabling complex audio envelope transformations with minimal computational overhead.
        """
        super().__init__(*args, **kwargs, **features)


    def process(
        self, 
        data: np.ndarray, 
        preset: Union[FadePreset, np.ndarray] = FadePreset.SMOOTH_ENDS, 
        **kwargs
    ) -> np.ndarray:
        """
        Shape your audio's dynamics with customizable envelope effects!

        This method allows you to apply various envelope shapes to your audio signal, 
        transforming its amplitude characteristics with precision and creativity. 
        Whether you want to smooth out transitions, create pulsing effects, 
        or craft unique fade patterns, this method has you covered.

        Parameters:
        -----------
        data : numpy.ndarray
            Your input audio data. Can be a single channel or multi-channel array.
            The envelope will be applied across the last dimension of the array.

        preset : FadePreset or numpy.ndarray, optional
            Define how you want to shape your audio's amplitude:
            
            - If you choose a FadePreset (default: SMOOTH_ENDS):
            Select from predefined envelope shapes like smooth fades, 
            bell curves, pulse effects, tremors, and more. Each preset 
            offers a unique way to sculpt your sound.
            
            - If you provide a custom numpy array:
            Create your own bespoke envelope by passing in a custom amplitude array. 
            This gives you ultimate flexibility in sound design.

        Additional keyword arguments (optional):
        ----------------------------------------
        Customize envelope generation with these powerful parameters:

        Envelope Generation Parameters:
        - enable_spline : bool
        Smoothen your envelope with spline interpolation. Great for creating 
        more organic, natural-feeling transitions.

        - spline_sigma : float, default varies
        Control the smoothness of spline interpolation. Lower values create 
        sharper transitions, higher values create more gradual blends.

        - fade_max_db : float, default 0.0
        Set the maximum amplitude in decibels. Useful for controlling peak loudness.

        - fade_max_min_db : float, default -60.0
        Define the minimum amplitude in decibels. Helps create subtle or dramatic fades.

        - fade_attack : float, optional
        Specify the proportion of the audio dedicated to the attack phase. 
        Influences how quickly the sound reaches its peak volume.

        - fade_release : float, optional
        Set the proportion of the audio dedicated to the release phase. 
        Controls how the sound tapers off.

        - buffer_size : int, default 400
        Adjust the internal buffer size for envelope generation.

        - sawtooth_freq : float, default 37.7
        For presets involving sawtooth wave modulation, control the frequency 
        of the underlying oscillation.

        Returns:
        --------
        numpy.ndarray
            Your processed audio data with the envelope applied.
            Maintains the same shape and type as the input data.

        Examples:
        ---------
            >>> from sudio.process.fx import FadeEnvelope, FadePreset
            >>> su = sudio.Master()        
            >>> rec = su.add('file.mp3')
            >>> rec.afx(FadeEnvelope, preset=FadePreset.PULSE, start=0, stop=10)
        """
        if isinstance(preset, (np.ndarray, list, tuple)):
            
            envelope = prepare_envelope(
            data.shape[-1], 
            np.array(preset, dtype=np.double), 
            **kwargs
            )
        elif isinstance(preset, FadePreset):
            envelope = generate_envelope(
                data.shape[-1], 
                preset.value, 
                **kwargs
                )
        else:
            raise TypeError('Invalid preset')
        
        processed_data = data * envelope
        return processed_data




class RefreshError(Exception):
    """Exception raised when a refresh operation encounters an error."""


class DecodeError(Exception):
    """Exception raised when an error occurs during audio decoding."""


class StreamError(Exception):
    """Exception raised for errors related to audio streaming."""




# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


import time

def convert_string_to_python_type(st):
    """
    Converts the input string 'st' to a more appropriate Python data type based on its content.

    Args:
        st (str): The input string to be converted.

    Returns:
        Union[int, float, dict, list, str]: The converted value. If 'st' represents a numeric value, it
        is converted to an integer or float. If 'st' contains curly braces, it is converted to a dictionary.
        If 'st' contains square brackets, it is converted to a list. Otherwise, the original string is returned.

    Example:
        >>> convert_string_to_python_type("42")
        42
        >>> convert_string_to_python_type("3.14")
        3.14
        >>> convert_string_to_python_type("{ 'key': 'value' }")
        {'key': 'value'}
        >>> convert_string_to_python_type("[1, 2, 3]")
        [1, 2, 3]
        >>> convert_string_to_python_type("Hello World")
        'Hello World'
    """
    if st.isnumeric():
        st = int(st)
    elif st.replace('.', '', 1).isnumeric():
        st = float(st)
    elif '{' in st and '}' in st:
        st = st.strip('{').strip('}')
        st = parse_dictionary_string(st)
    elif '[' in st and ']' in st:
        st = st.strip('[').strip(']')
        st = parse_list_string(st)
    return st


def parse_dictionary_string(st, dict_eq=':', item_sep=','):
    """
    Converts a string 'st' representing a dictionary-like structure to a Python dictionary.

    Args:
        st (str): The input string representing the dictionary.
        dict_eq (str, optional): The string used as the key-value separator. Defaults to ':'.
        item_sep (str, optional): The string used as the item separator. Defaults to ','.

    Returns:
        dict: The converted dictionary.

    Example:
        >>> parse_dictionary_string("{key1:value1, key2:value2}")
        {'key1': 'value1', 'key2': 'value2'}
        >>> parse_dictionary_string("name:John, age:30, city:New York", dict_eq=':', item_sep=', ')
        {'name': 'John', 'age': 30, 'city': 'New York'}
        >>> parse_dictionary_string("a=1; b=2; c=3", dict_eq='=', item_sep='; ')
        {'a': 1, 'b': 2, 'c': 3}
    """
    if not st:
        return None
    buff = {}
    i = [i.strip().split(dict_eq) for i in st.split(item_sep)]
    for j in i:
        if j:
            value = j[1] if len(j) > 1 else None
            if value:
                value = convert_string_to_python_type(value)
            buff[j[0]] = value
    return buff


def parse_list_string(st):
    """
    Converts a string 'st' representing a list-like structure to a Python list.

    Args:
        st (str): The input string representing the list.

    Returns:
        list: The converted list.

    Example:
        >>> parse_list_string("[1, 2, 3, 4, 5]")
        [1, 2, 3, 4, 5]
        >>> parse_list_string("apple, orange, banana, mango")
        ['apple', 'orange', 'banana', 'mango']
        >>> parse_list_string("3.14, 2.71, 1.618", convert_string_to_python_type=float)
        [3.14, 2.71, 1.618]
    """
    if not st:
        return None
    buff = []
    i = [i.strip() for i in st.split(',')]
    for value in i:
        if value:
            value = convert_string_to_python_type(value)
        buff.append(value)
    return buff



def generate_timestamp_name(seed=None):
    """
    Generates a timestamp-based name.

    Args:
        seed (str, optional): Seed value for additional uniqueness. Defaults to None.

    Returns:
        str: The generated timestamp-based name.

    Example:
        >>> generate_timestamp_name()
        'YYYYMMDD_SSMMM'
        >>> generate_timestamp_name("custom_seed")
        'custom_seedDD_SSMMM'

    Note:
        - 'YYYY' represents the year.
        - 'MM' represents the month.
        - 'DD' represents the day.
        - 'SS' represents the second.
        - 'MMM' represents the millisecond.
        - The seed, if provided, is appended to the beginning of the name.
    """
    current_time = time.gmtime()
    if seed:
        name = str(seed) + str(current_time.tm_mday) + str(current_time.tm_sec) + str(time.time() % 1)[2:5]
    else:
        name = (str(current_time.tm_year) + str(current_time.tm_mon) +
                str(current_time.tm_mday) + '_' + str(current_time.tm_sec) + str(time.time() % 1)[2:5])
    return name





# cython: boundscheck=False, wraparound=False, nonecheck=False, cdivision=True

# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


cimport numpy as np

cpdef np.ndarray pitch_shifter_cy(
        np.ndarray input_audio, 
        np.ndarray[double, ndim=1] envelope,
        float ratio=*,
        int sample_rate=*,
        bint enable_spline=*,
        float spline_sigma =*,
        float fade_max_db =*,
        float fade_min_db =*,
        int envbuffer =*,
        int envlen =*,
        int frame_length =*,
        int sequence_ms =*,
        int seekwindow_ms =*,
        int overlap_ms =*,
        object converter_type=*,
)


cpdef np.ndarray _pitch_shifter_cy(
    np.ndarray input_audio, 
    object intp,
    int sample_rate=*,
    int frame_length =*,
    int sequence_ms =*,
    int seekwindow_ms =*,
    int overlap_ms =*,
    object converter_type=*,
)




# cython: language_level=3
# distutils: extra_compile_args = -O3


# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


import numpy as np
cimport numpy as np
cimport cython


cpdef int find_nearest_divisible(int reference_number, int divisor):
    """
    Finds the number closest to 'reference_number' that is divisible by 'divisor'.

    Args:
        reference_number (int): The reference number.
        divisor (int): The divisor.
    """
    cdef int modulo = reference_number % divisor
    cdef int quotient = reference_number // divisor
    cdef int option1 = reference_number - modulo
    cdef int option2 = reference_number + (divisor - modulo)

    return option1 if abs(option1 - reference_number) <= abs(option2 - reference_number) else option2

cpdef int find_nearest_divisor(int num, int divisor) except? -1:
    """
    Finds the nearest divisor with zero remainder for 'num'.

    Args:
        num (int): The dividend.
        divisor (int): The candidate divisor.

    Returns:
        int: The nearest divisor.
    """
    cdef int div = int(round(num / divisor))
    cdef int lower = div
    cdef int upper = div
    
    while upper < num:
        if num % lower == 0:
            return lower
        if num % upper == 0:
            return upper
        
        lower -= 1
        upper += 1
    
    raise ValueError("No divisor with a zero remainder found.")



@cython.boundscheck(False)
@cython.wraparound(False)
cpdef db2amp(db):
    """
    Convert decibels to amplitude.

    Args:
        db (int, float, ndarray): Decibel value(s)

    """
    return np.power(10.0, (db / 20.0), dtype=np.float64)

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef amp2db(amp):
    """
    Convert amplitude to decibels.

    Args:
        amp (int, float, ndarray): Amplitude value(s)

    """
    return 20.0 * np.log10(amp, dtype=np.float64)

    



# SUDIO - Audio Processing Platform
# Copyright (C) 2024 Hossein Zahaki

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
#  any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# - GitHub: https://github.com/MrZahaki/sudio


import scipy.signal as scisig
import threading
import queue
import numpy as np
from typing import Union, Callable
import warnings
import gc
import time
import os
from io import BufferedRandom
import traceback
from pathlib import Path


from sudio.types import StreamMode, RefreshError
from sudio.types import PipelineProcessType
from sudio.process.audio_wrap import AudioWrap
from sudio.utils.strtool import generate_timestamp_name 
from sudio.utils.timed_indexed_string import TimedIndexedString 
from sudio.stream.stream import Stream
from sudio.stream.streamcontrol import StreamControl
from sudio.utils.window import multi_channel_overlap, single_channel_overlap
from sudio.utils.window import multi_channel_windowing, single_channel_windowing
from sudio.utils.channel import shuffle3d_channels, shuffle2d_channels, get_mute_mode_data, map_channels
from sudio.audiosys.sync import synchronize_audio
from sudio.audiosys.cacheman import handle_cached_record, write_to_cached_file
from sudio.pipeline import Pipeline
from sudio.metadata import AudioRecordDatabase, AudioMetadata
from sudio.io import SampleFormat, codec, write_to_default_output, AudioStream
from sudio.io import AudioDeviceInfo, get_sample_size, FileFormat
from sudio.io._webio import WebAudioIO



class Master:

    CACHE_INFO_SIZE = 4 * 8
    BUFFER_TYPE = '.su'

    def __init__(self,
                 input_device_index: int = None,
                 output_device_index: int = None,
                 data_format: SampleFormat = SampleFormat.SIGNED16,
                 nperseg: int = 500,
                 noverlap: int = None,
                 window: object = 'hann',
                 NOLA_check: bool = True,
                 input_dev_sample_rate: int = 44100,
                 input_dev_nchannels: int = 2,
                 input_dev_callback: Callable = None,
                 output_dev_nchannels:int = 2,
                 output_dev_callback: Callable = None,
                 buffer_size: int = 30,
                 audio_data_directory: str = './sudio/',
                 ):
        
        """
        The `Master` class is responsible for managing audio data streams, applying windowing, 
        and handling input/output devices. This class provides various methods for processing, 
        recording, and playing audio data, with functionality to handle different audio formats, 
        sample rates, and device configurations.


		Parameters
        ----------
        - input_device_index : int, optional
            Index of the input audio device. If None, uses the system's default input device.
        - output_device_index : int, optional
            Index of the output audio device. If None, uses the system's default output device.
        - data_format : SampleFormat, default=SampleFormat.SIGNED16
            Audio sample format (e.g., FLOAT32, SIGNED16, UNSIGNED8)
        - nperseg : int, default=500
            Number of samples per segment for windowing and processing
        - noverlap : int, optional
            Number of overlapping samples between segments. 
            If None, defaults to half of nperseg.
        - window : str, float, tuple, or ndarray, default='hann'
            Window function type for signal processing:

            - String: scipy.signal window type (e.g., 'hamming', 'blackman')
            - Float: Beta parameter for Kaiser window
            - Tuple: Window name with parameters
            - ndarray: Custom window values
            - None: Disables windowing

        - NOLA_check : bool, default=True
            Perform Non-Overlap-Add (NOLA) constraint verification
        - input_dev_sample_rate : int, default=44100
            Input device sample rate in Hz. Behavior depends on input:

            - If a specific value is provided: Uses the given sample rate
            - If None: Automatically selects the default sample rate of the input device(if it exist)
            - If the selected rate is unsupported, raises an error
            - Recommended range typically between 8000 Hz and 96000 Hz

        - input_dev_nchannels : int, default=2
            Number of input device channels
        - input_dev_callback : callable, optional
            Custom callback for input device processing
        - output_dev_nchannels : int, default=2
            Number of output device channels
        - output_dev_callback : callable, optional
            Custom callback for output device processing
        - buffer_size : int, default=30
            Size of the audio stream buffer
        - audio_data_directory : str, default='./sudio/'
            Directory for storing audio data files

        Notes
        -----
        
        - Various methods are available for audio processing, pipeline management, and device control.
          Refer to individual method docstrings for details.
        - The class uses multi-threading for efficient audio stream management.
        - Window functions are crucial for spectral analysis and should be chosen carefully.
        - NOLA constraint ensures proper reconstruction in overlap-add methods.
        - Custom callbacks allow for flexible input/output handling but require careful implementation.
        """

        self._sound_buffer_size = buffer_size
        self._stream_type = StreamMode.optimized
        self._sample_format_type = data_format
        self.input_dev_callback = None
        self.output_dev_callback = None
        self._default_stream_callback = self._input_stream_callback
        self._sample_format = data_format
        self._nperseg = nperseg
        self._data_chunk = nperseg
        self._sample_width = get_sample_size(data_format)
        self._sample_width_format_str = (
            '<f{}'.format(self._sample_width) if data_format == SampleFormat.FLOAT32.value else '<i{}'.format(self._sample_width)
        )

        self._exstream_mode = threading.Event()
        self._master_mute_mode = threading.Event()
        self._stream_loop_mode = False
        self._stream_on_stop = False
        self._stream_data_pointer = 0
        self._stream_data_size = None
        self._stream_file = None
        self._nchannels = None
        self._sample_rate = None
        self._audio_data_directory = audio_data_directory
        try:
            os.mkdir(self._audio_data_directory)
        except FileExistsError:
            pass

        self._output_device_index = None
        input_channels = int(1e6)
        output_channels = int(1e6)

        if input_dev_sample_rate is not None:
            assert isinstance(input_dev_sample_rate, int), TypeError('control input_dev_sample_rate')

        try:
            if callable(input_dev_callback):
                self.input_dev_callback = input_dev_callback
                self._default_stream_callback = self._custom_stream_callback

            try:
                count = self.get_device_count()
            except Exception as e:
                raise EOFError(f'{e}')
            if count < 1: raise EOFError('No audio input devices were found')
        
            if input_device_index is None:
                try:
                    dev = self.get_default_input_device_info()
                except:
                    raise EOFError(f'No default input info or device')
                
            else:
                try:
                    int(input_device_index)
                except ValueError:
                    raise ValueError('invalid literal for input_device_index.')
                
                try:
                    dev = self.get_device_info_by_index(input_device_index)
                except:
                    raise EOFError(f'No input info or device for index {input_device_index}')

            try:
                self._sample_rate = int(dev.default_sample_rate) if input_dev_sample_rate is None else input_dev_sample_rate
                assert self._sample_rate > 0.0
            except:
                raise EOFError(f'Input device samplerate unsupported')
            
            try:
                input_channels = int(dev.max_input_channels)
                assert input_channels > 0
            except:
                raise EOFError(f'Input device channels unsupported')
            
            try:
                self._input_device_index = int(dev.index)

                audio_input_stream = AudioStream()
                try:
                    audio_input_stream.open(
                        input_dev_index = self._input_device_index,
                        sample_rate = self._sample_rate,
                        format=self._sample_format,
                        input_channels=input_channels,
                        frames_per_buffer=self._data_chunk,
                        enable_input=True,
                        enable_output=False,
                        ) 
                finally:
                    audio_input_stream.close()
                
            except Exception as e:
                raise EOFError(f'Input device of {self._input_device_index}  {e}')


        except EOFError as e:
            warnings.warn(f"{e}. some features will be unavailable.")
            self._input_device_index = None

            try:
                input_channels = int(input_dev_nchannels)  
                assert input_channels > 0
            except:
                raise ValueError('Unsupported input dev channels.')
            
            try:
                self._sample_rate = int(input_dev_sample_rate)  # Default sample rate
                assert self._sample_rate > 0
            except:
                raise ValueError('Unsupported samplerate.')


        try:
            if callable(output_dev_callback):
                self.output_dev_callback = output_dev_callback
            try:
                count = self.get_device_count()
            except Exception as e:
                raise EOFError(f'{e}')
            if count < 1: raise EOFError('No audio output devices were found')
            
            if output_device_index is None:
                try:
                    dev = self.get_default_output_device_info()
                except:
                    raise EOFError(f'No default output info or device')
                
            else:
                try:
                    int(output_device_index)
                except ValueError:
                    raise ValueError('invalid literal for output_device_index.')
                    
                try:
                    dev = self.get_device_info_by_index(output_device_index)
                except:
                    raise EOFError(f'No output info or device for index {output_device_index}')

            try:
                output_channels = dev.max_output_channels
                assert output_channels > 0
            except:
                raise EOFError(f'Output device channels unsupported')
            try:
                self._output_device_index = dev.index

                audio_output_stream = AudioStream()
                try:
                    audio_output_stream.open(
                        output_dev_index = self._output_device_index,
                        sample_rate = self._sample_rate,
                        format=self._sample_format,
                        output_channels=output_channels,
                        frames_per_buffer=self._data_chunk,
                        enable_input=False,
                        enable_output=True,
                        ) 
                finally:
                    audio_output_stream.close()
                    
            except Exception as e:
                raise EOFError(f'Output device of {self._output_device_index} {e}')

        except EOFError as e:
            warnings.warn(f"{e}. some features will be unavailable.")
            self._output_device_index = None

            try:
                output_channels = int(output_dev_nchannels)  
                assert output_channels > 0
            except:
                raise ValueError('Unsupported output dev channels.')

        self._input_channels = input_channels
        self._output_channels = output_channels
        self._nchannels = self._output_channels

        if self._nchannels == 0:
            raise ValueError('No input or output device found')
        self.mute()
        self.branch_pipe_database_min_key = 0
        self._window_type = window

        if noverlap is None:
            noverlap = nperseg // 2
        if type(self._window_type) is str or type(self._window_type) is tuple or type(self._window_type) == float:
            window = scisig.get_window(window, nperseg)
        if self._window_type:
            assert len(window) == nperseg, 'Control size of window'
            if NOLA_check:
                assert scisig.check_NOLA(window, nperseg, noverlap)
        elif self._window_type is None:
            window = None
            self._window_type = None

        self._threads = []
        self._local_database = AudioRecordDatabase()
        self._recordq = [threading.Event(), queue.Queue(maxsize=0)]
        self._queue = queue.Queue()
        self._echo_flag = threading.Event()
        self.main_pipe_database = []
        self.branch_pipe_database = []
        self._functions = []
        for i in range(len(self._functions)):
            self._threads.append(threading.Thread(target=self._run, daemon=True, args=(len(self._threads),)))
            self._queue.put(i)
        # Set parameters for windowing
        self._nhop = nperseg - noverlap
        self._noverlap = noverlap
        self._window = window
        self._overlap_buffer = [np.zeros(nperseg) for i in range(self._nchannels)]
        self._windowing_buffer = [[np.zeros(nperseg), np.zeros(nperseg)] for i in range(self._nchannels)]
        data_queue = queue.Queue(maxsize=self._sound_buffer_size)
        self._normal_stream = Stream(self, data_queue, process_type=PipelineProcessType.QUEUE)
        self._main_stream = Stream(self, data_queue, process_type=PipelineProcessType.QUEUE)
        self._refresh_ev = threading.Event()
        # Master selection and synchronization
        self._master_sync = threading.Barrier(self._nchannels)
        self._audio_input_stream = False
        self._audio_output_stream = False
        self.clean_cache()


    def start(
            self,
            ):
        """
        Starts the audio input and output streams and launches any registered threads.

        Returns
        -------

        self : Master
            Returns the instance of the Master class for method chaining.
        """
        assert self._output_device_index, "No output device, streaming unavailable"
        assert self._input_device_index, "No input device, streaming unavailable"
        assert not self._audio_input_stream, 'Master is Already Started'


        try:
            self._audio_input_stream = AudioStream()
            self._audio_output_stream = AudioStream()

            self._audio_input_stream.open(
                input_dev_index = self._input_device_index,
                sample_rate = self._sample_rate,
                format=self._sample_format,
                input_channels=self._input_channels,
                frames_per_buffer=self._data_chunk,
                enable_input=True,
                enable_output=False,
                input_callback=self._default_stream_callback
                )
            
            self._audio_output_stream.open(
                    output_dev_index=self._output_device_index,
                    frames_per_buffer=self._data_chunk,
                    sample_rate=self._sample_rate, 
                    format=self._sample_format,
                    output_channels=self._output_channels,
                    enable_input=False, 
                    enable_output=True,
                    output_callback=self._output_stream_callback
                )

            self._audio_output_stream.start()
            self._audio_input_stream.start()
            
            for i in self._threads:
                i.start()


        except:
            raise

    def _run(self, th_id):
        """
        Grabs a function from the queue and runs it with the given thread ID, 
        then marks the task as complete. Essentially a simple worker method 
        to keep things moving smoothly.
        """
        self._functions[self._queue.get()](th_id)
        self._queue.task_done()


    def _exstream(self):
        """
        Handles external stream file reading and processing.

        Reads audio data chunk by chunk from a file, manages looping or 
        stopping behavior, and puts the data into the main stream. Handles 
        edge cases like reaching end of file or needing to loop back.
        """
        while self._exstream_mode.is_set():
            in_data: np.ndarray = np.frombuffer(
                self._stream_file.read(self._stream_data_size),
                self._sample_width_format_str).astype('f')
            try:
                in_data: np.ndarray = map_channels(in_data, self._nchannels, self._nchannels)
            except ValueError:
                return

            if not in_data.shape[-1] == self._data_chunk:
                if self._stream_loop_mode:
                    self._stream_file.seek(self._stream_data_pointer, 0)
                else:
                    self._exstream_mode.clear()
                    self._stream_file.seek(self._stream_data_pointer, 0)
                    self._main_stream.clear()

                    self._stream_file = None
                    self._stream_loop_mode = None
                    self._stream_data_pointer = None
                    self._stream_data_size = None

                if self._stream_on_stop:
                    self._stream_on_stop()
                    self._stream_on_stop = None
                return
            
            self._main_stream.acquire()
            self._main_stream.put(in_data)  
            self._main_stream.release()
            if self._stream_type == StreamMode.normal:
                break
    
    def _main_stream_safe_release(self):
        """
        Safely releases the main stream lock if it's currently locked.

        Prevents potential deadlocks by checking and releasing the stream 
        lock only if it's actually held. A little insurance policy for 
        thread safety.
        """
        if self._main_stream.locked():
            self._main_stream.release()


    def _input_stream_callback(self, in_data, frame_count, format):  
        """
        Processes incoming audio input stream data.

        Handles different modes like external streaming, mute, and normal 
        input. Manages channel mapping, puts data into the main stream, and 
        handles potential errors gracefully. The Swiss Army knife of 
        audio input processing.
        """

        if self._exstream_mode.is_set():
            self._exstream()
            return None, True
        
        elif self._master_mute_mode.is_set():
            in_data = get_mute_mode_data(self._nchannels, self._nperseg)
        else:
            in_data = np.frombuffer(in_data, self._sample_width_format_str).astype('f')
            try:
                in_data = map_channels(in_data, self._input_channels, self._nchannels)
            except Exception as e:
                error_msg = f"Error in audio channel mapping: {type(e).__name__}: {str(e)}"
                print(error_msg)
                print(traceback.format_exc())

                return None, True
        try:
            self._main_stream.acquire()
            self._main_stream.put(in_data)  
            self._main_stream.release()

        except Exception as e:
            # print(f"Error in stream callback: {e}")
            # print(f"in_data shape: {in_data.shape}, type: {type(in_data)}")
            # print(f"self._input_channels: {self._input_channels}, self._nchannels: {self._nchannels}")
            # print(traceback.format_exc())  # This will print the full stack trace
            return None, False  # Indicate that an error occurred

        return None, True


    def _custom_stream_callback(self, in_data, frame_count, format):  

        if self._exstream_mode.is_set():
            if not self._exstream():
                return None, 0

        elif self._master_mute_mode.is_set():
            in_data = get_mute_mode_data(self._nchannels, self._nperseg)
        else:
            in_data = self.input_dev_callback(frame_count, format).astype('f')
            try:
                in_data = map_channels(in_data, self._input_channels, self._nchannels)
            except Exception as e:
                error_msg = f"Error in audio channel mapping: {type(e).__name__}: {str(e)}"
                print(error_msg)
                print(traceback.format_exc())  # This will print the full stack trace

        try:
            self._main_stream.acquire()
            self._main_stream.put(in_data)  
            self._main_stream.release()

        except Exception as e:
            print(f"Error in stream callback: {e}")
            print(f"in_data shape: {in_data.shape}, type: {type(in_data)}")
            print(f"self._input_channels: {self._input_channels}, self._nchannels: {self._nchannels}")
            print(traceback.format_exc())  # This will print the full stack trace
            return None, 1  # Indicate that an error occurred

        # self.a1 = time.perf_counter()
        return None, 0

    def _output_stream_callback(self, frame_count, format):
        rec_ev, rec_queue = self._recordq
        while 1:
            data = None

            try:
                data = self._main_stream.get(timeout=0.0001)

                if rec_ev.is_set():
                    rec_queue.put_nowait(data)
            except:
                pass

            try:
                if data is not None:
                    data = shuffle2d_channels(data)

                    if self.output_dev_callback:
                        self.output_dev_callback(data)
                    
                    if self._echo_flag.is_set():
                        return (data.tobytes(), True)

            except Exception as e:
                error_msg = f"Error in _output_stream_callback: {type(e).__name__}: {str(e)}"
                print(error_msg)
                print(traceback.format_exc())  

            if self._refresh_ev.is_set():
                self._refresh_ev.clear()
                # close current stream
                try:
                    self._audio_output_stream.stop()
                    self._audio_output_stream.close()
                except:
                    pass
                    # print(f"Error closing stream: {type(e).__name__}: {str(e)}")
                    # print(traceback.format_exc())    
                            
                # create new stream
                try:
                    self._audio_output_stream.open(
                        output_dev_index=self._output_device_index,
                        frames_per_buffer=self._data_chunk,
                        sample_rate=self._sample_rate, 
                        format=self._sample_format,
                        output_channels=self._output_channels,
                        enable_input=False, 
                        enable_output=True,
                        output_callback=self._output_stream_callback
                    )
                    self._audio_output_stream.start()

                except Exception as e:
                    print(f"Error opening new stream: {type(e).__name__}: {str(e)}")  
                    print(traceback.format_exc())                     

                self._main_stream.clear()


    def add_file(self, filename: str, sample_format: SampleFormat = SampleFormat.UNKNOWN,
                nchannels: int = None, sample_rate: int = None ,safe_load=True):
        '''
        Adds an audio file to the database with optional format and parameter adjustments.

        Supports WAV, FLAC, VORBIS, and MP3 file formats.

        Parameters:
        -----------

        filename : str
            Path/name of the audio file to add.
        sample_format : SampleFormat, optional
            Desired sample format. Defaults to automatically detecting or using master's format.
        nchannels : int, optional
            Number of audio channels. Defaults to file's original channel count.
        sample_rate : int, optional
            Desired sample rate. Defaults to file's original rate.
        safe_load : bool, default True
            If True, modifies file to match master object's audio attributes.

        Returns:
        --------

        AudioWrap
            Wrapped audio file with processed metadata and data.

        Raises:
        -------

        ImportError
            If safe_load is True and file's channel count exceeds master's channels.

        '''
        info = codec.get_file_info(filename)
        if safe_load:
            sample_format = self._sample_format
        elif sample_format is SampleFormat.UNKNOWN:
            if info.sample_format is SampleFormat.UNKNOWN:
                sample_format = self._sample_format
            else:
                sample_format = info.sample_format
        else:
            sample_format = sample_format

        if nchannels is None:
            nchannels = info.nchannels
        if safe_load:
          sample_rate = self._sample_rate
        elif sample_rate is None:
            sample_rate = info.sample_rate

        p0 = max(filename.rfind('\\'), filename.rfind('/')) + 1
        p1 = filename.rfind('.')
        if p0 < 0:
            p0 = 0
        if p1 <= 0:
            p1 = None
        name = filename[p0: p1]
        if name in self._local_database.index():
            name = name + generate_timestamp_name()

        record = AudioMetadata(name, **{
                'size': None,
                'noise': None,
                'frameRate': sample_rate,
                'o': None,
                'sampleFormat': sample_format,
                'nchannels': nchannels,
                'duration': info.duration,
                'nperseg': self._nperseg,
            }
        )

        if safe_load and  record.nchannels > self._nchannels:
            raise ImportError('number of channel for the {name}({ch0})'
                              ' is not same as object channels({ch1})'.format(name=name,
                                                                              ch0=record.nchannels,
                                                                              ch1=self._nchannels))
        decoder = lambda: codec.decode_audio_file(filename, sample_format, nchannels, sample_rate)
        path_server = TimedIndexedString(
                self._audio_data_directory + name + Master.BUFFER_TYPE,
                start_before=Master.BUFFER_TYPE
                )
        orphaned_cache = self.prune_cache()

        record = handle_cached_record(
            record=record,
            path_server=path_server,
            orphaned_cache=orphaned_cache,
            cache_info_size=self.__class__.CACHE_INFO_SIZE,
            decoder=decoder,
            sync_sample_format_id=self._sample_format,
            sync_nchannels=self._nchannels,
            sync_sample_rate=self._sample_rate,
            safe_load = safe_load,
            )
        self._local_database.add_record(record)
        gc.collect()
        return self.load(name)


    def add(self, record, safe_load=True):
        """
        Adds audio data to the local database from various input types.

        Supports adding:
        - AudioWrap objects
        - Audio file paths (mp3, WAV, FLAC, VORBIS)
        - AudioMetadata records

        Parameters:
        -----------

        record : Union[AudioWrap, str, AudioMetadata]
            The audio record to add to the database.
        safe_load : bool, default True
            If True, ensures record matches master object's audio specifications.

        Returns:
        --------

        AudioWrap
            Processed and wrapped audio record.

        Raises:
        -------

        ImportError
            If safe_load is True and record's channel count exceeds master's.
        TypeError
            If record is not a supported type.

        Notes:
        ------

        - Uses cached files for optimized memory management
        - Execution time may vary based on cached file state

        Examples:
        ---------

        .. code-block:: python
            
            master = sudio.Master()
            audio = master.add('./alan kujay.mp3')

            master1 = sudio.Master()
            audio1 = master1.add(audio)

            master.echo(audio[:10])
            master1.echo(audio1[:10])
        """
        if isinstance(record, AudioWrap):
            record = record.get_data()
            return self.add(record, safe_load=safe_load)

        elif isinstance(record, AudioMetadata):
            name = record.name
            if safe_load and record.nchannels > self._nchannels:
                raise ImportError('number of channel for the {name}({ch0})'
                                ' is not same as object channels({ch1})'.format(name=name,
                                                                                ch0=record.nchannels,
                                                                                ch1=self._nchannels))
            if type(record.o) is not BufferedRandom:
                if record.name in self._local_database.index():
                    record.name = generate_timestamp_name()

                if safe_load:
                    record = self._sync_record(record)

                f, newsize = (
                    write_to_cached_file(record.size,
                                record.frameRate,
                                record.sampleFormat if record.sampleFormat else 0,
                                record.nchannels,
                                file_name=self._audio_data_directory + record.name + Master.BUFFER_TYPE,
                                data=record.o,
                                pre_truncate=True,
                                after_seek=(Master.CACHE_INFO_SIZE, 0),
                                after_flush=True,
                                size_on_output=True)
                )
                record.o = f
                record.size = newsize

            elif (not (self._audio_data_directory + record.name + Master.BUFFER_TYPE) == record.name) or \
                    record.name in self._local_database.index():
                if record.name in self._local_database.index():
                    record.name = generate_timestamp_name()
                
                prefile = record.o
                prepos = prefile.tell(), 0

                record.o, newsize = (
                    write_to_cached_file(record.size,
                                record.frameRate,
                                record.sampleFormat if record.sampleFormat else 0,
                                record.nchannels,
                                file_name=self._audio_data_directory + record.name + Master.BUFFER_TYPE,
                                data=prefile.read(),
                                after_seek=prepos,
                                after_flush=True,
                                size_on_output=True)
                )
                prefile.seek(*prepos)
                record.size = newsize
            self._local_database.add_record(record)

            gc.collect()
            return self.wrap(record)

        elif isinstance(record, str):
            gc.collect()
            return self.add_file(record, safe_load=safe_load)

        else:
            raise TypeError('The record must be an audio file, data frame or a AudioWrap object')


    def recorder(self, record_duration: float, name: str = None):
        """
        Record audio for a specified duration. 
        This method captures audio input for a given duration and stores it as a new record
        in the Master object's database.

        Parameters:
        -----------

        record_duration : float
            The duration of the recording in seconds.

        name : str, optional
            A custom name for the recorded audio. If None, a timestamp-based name is generated.

        :return: AudioWrap instance


        Notes:
        ------

        - The recording uses the current audio input settings of the Master object
         (sample rate, number of channels, etc.).
        - The recorded audio is automatically added to the Master's database and can be
         accessed later using the provided or generated name.
        - This method temporarily modifies the internal state of the Master object to
         facilitate recording. It restores the previous state after recording is complete.

        Examples:
        ---------

        **Record for 5 seconds with an auto-generated name**

         >>> recorded_audio = master.recorder(5)

        **Record for 10 seconds with a custom name**

         >>> recorded_audio = master.recorder(10, name="my_recording")

        **Use the recorded audio**

         >>> master.echo(recorded_audio)
        """

        assert self.is_started(), "instance not started, use start()"

        if name is None:
            name = generate_timestamp_name('record')
        elif name in self._local_database.index():
            raise KeyError(f'The name "{name}" is already registered in the database.')

        rec_ev, rec_queue = self._recordq
        rec_ev.set()  # Start recording

        # Record for the specified duration
        start_time = time.time()
        recorded_data = []
        while time.time() - start_time < record_duration:
            data = rec_queue.get()
            recorded_data.append(data)

        rec_ev.clear()  # Stop recording

        sample = np.array(recorded_data)
        sample = shuffle3d_channels(sample)

        metadata = AudioMetadata(
            name, 
            **{
                'size': sample.nbytes,
                'noise': None,
                'frameRate': self._sample_rate,
                'nchannels': self._nchannels,
                'sampleFormat': self._sample_format,
                'nperseg': self._nperseg,
                'duration': record_duration,
            }
        )

        metadata.o = write_to_cached_file(
            metadata.size,
            metadata.frameRate,
            metadata.sampleFormat,
            metadata.nchannels,
            file_name=f"{self._audio_data_directory}{name}{Master.BUFFER_TYPE}",
            data=sample.tobytes(),
            pre_truncate=True,
            after_seek=(Master.CACHE_INFO_SIZE, 0),
            after_flush=True
        )

        metadata.size += Master.CACHE_INFO_SIZE
        self._local_database.add_record(metadata)
        return self.wrap(metadata.copy())


    def load(self, name: str, safe_load: bool=True,
         series: bool=False) -> Union[AudioWrap, AudioMetadata]:
        '''
        Loads a record from the local database. Trying to load a record that was previously loaded, 
        outputs a wrapped version of the named record.

        :param name: record name
        :param safe_load: Flag to safely load the record. if safe load is enabled then load function tries to load a record
         in the local database based on the master settings, like the frame rate and etc (default: `True`).
        :param series:  Return the record as a series (default: `False`).
        :return: (optional) AudioWrap object, AudioMetadata
        
        '''

        if name in self._local_database.index():
            rec = self._local_database.get_record(name).copy()
            if series:
                return rec
            return self.wrap(rec)
        else:
            raise ValueError('can not found the {name} in database'.format(name=name))

        if safe_load and not self._mono_mode and rec['nchannels'] > self._nchannels:
            raise ImportError('number of channel for the {name}({ch0})'
                            ' is not same as object channels({ch1})'.format(name=name,
                                                                            ch0=rec['nchannels'],
                                                                            ch1=self._nchannels))
        if safe_load:
            rec['o'] = file.read()
            rec = self._sync_record(rec)
            rec['o'] = file

        if file_size > rec['size']:
            file.seek(rec['size'], 0)
        else:
            file.seek(Master.CACHE_INFO_SIZE, 0)

        self._local_database.add_record(rec)
        if series:
            return rec
        return self.wrap(rec)
    

    def get_record_info(self, record: Union[str, AudioWrap]) -> dict:
        '''
        Retrieves metadata for a given record.

        :param record: The record (str, or AudioWrap) whose info is requested.

        :return: information about saved record in a dict format ['frameRate'  'sizeInByte' 'duration'
            'nchannels' 'nperseg' 'name'].
        '''
        if isinstance(record, AudioWrap):
            name = record.name
        elif isinstance(record, str):
            name = record
        else:
            raise TypeError('record must be an instance of AudioWrap or str')
        
        if name in self._local_database.index():
            rec = self._local_database.get_record(name)

        else:
            raise ValueError('can not found the {name} in database'.format(name=name))
        return {
            'frameRate': rec['frameRate'],
            'sizeInByte': rec['size'],
            'duration': rec['duration'],
            'nchannels': rec['nchannels'],
            'nperseg': rec['nperseg'],
            'name': name,
            'sampleFormat': rec['sampleFormat'].name
        }

    def _syncable(self,
                 *target,
                 nchannels: int = None,
                 sample_rate: int = None,
                 sample_format_id: int = None):
        '''
         Determines whether the target can be synced with specified properties or not

        :param target: wrapped objects
        :param nchannels: number of channels; if the value is None, the target will be compared to the 'self' properties.
        :param sample_rate: sample rate; if the value is None, the target will be compared to the 'self' properties.
        :param sample_format_id: if the value is None, the target will be compared to the 'self' properties.


        :return: only objects that need to be synchronized.
        '''
        nchannels = nchannels if nchannels else self._nchannels
        sample_format_id = self._sample_format if sample_format_id is None else sample_format_id
        sample_rate = sample_rate if sample_rate else self._sample_rate

        buffer = []
        for rec in target:
            assert isinstance(rec, AudioWrap)
            tmp = rec.get_data()

            if (not tmp['nchannels'] == nchannels or
                not sample_rate == tmp['frameRate'] or
                not tmp['sampleFormat'] == sample_format_id):
                buffer.append(True)
            else:
                buffer.append(False)
        if len(buffer) == 1:
            return buffer[0]
        return buffer


    def syncable(self,
                 *target,
                 nchannels: int = None,
                 sample_rate: int = None,
                 sample_format: SampleFormat = SampleFormat.UNKNOWN):
        '''
         Prepares a list of targets to be synchronized. Determines whether the target can be synced with specified properties or not

        :param target: Targets to sync. wrapped objects
        :param nchannels: Number of channels (default: `None`); if the value is None, the target will be compared to the 'self' properties.
        :param sample_rate: Sample rate (default: `None`); if the value is None, the target will be compared to the 'self' properties.
        :param sample_format: Sample format (default: `SampleFormat.UNKNOWN`); if the value is None, the target will be compared to the 'self' properties.

        :return: only objects that need to be synchronized.
        '''
        return self._syncable(*target, nchannels=nchannels, sample_rate=sample_rate, sample_format_id=sample_format)


    def sync(self,
            *targets,
            nchannels: int=None,
            sample_rate: int=None,
            sample_format: SampleFormat=SampleFormat.UNKNOWN,
            output='wrapped'):
        '''
        Synchronizes audio across multiple records. Synchronizes targets in the AudioWrap object format with the specified properties.

        :param targets: Records to sync. wrapped objects.
        :param nchannels: Number of channels (default: `None`); if the value is None, the target will be synced to the 'self' properties.
        :param sample_rate: Sample rate (default: `None`); if the value is None, the target will be synced to the 'self' properties.
        :param sample_format: if the value is None, the target will be synced to the 'self' properties.
        :param output: can be 'wrapped', 'series' or 'ndarray_data'

        :return: synchronized objects.
        '''
        nchannels = nchannels if nchannels else self._nchannels
        sample_format = self._sample_format if sample_format == SampleFormat.UNKNOWN else sample_format
        sample_rate = sample_rate if sample_rate else self._sample_rate

        out_type = 'ndarray' if output.startswith('n') else 'byte'

        buffer = []
        for record in targets:
            record: AudioMetadata = record.get_data().copy()
            assert isinstance(record, AudioMetadata)
            main_file: BufferedRandom = record.o
            main_seek = main_file.tell(), 0
            record.o = main_file.read()
            main_file.seek(*main_seek)
            synchronize_audio(record, nchannels, sample_rate, sample_format, output_data=out_type)
            record.name = generate_timestamp_name(record.name)
            if output[0] == 'w':
                buffer.append(self.add(record))
            else:
                buffer.append(record)
        return tuple(buffer)


    def _sync_record(self, rec):
        return synchronize_audio(rec, self._nchannels, self._sample_rate, self._sample_format)

    def del_record(self, record: Union[str, AudioMetadata, AudioWrap]):
        '''
        Deletes a record from the local database.

        :param record: Record to delete (str, AudioMetadata, or AudioWrap).
        '''

        if isinstance(record, (AudioMetadata, AudioWrap)):
            name = record.name
        elif isinstance(record, str):
            name = record
        else:
            raise TypeError('please control the type of record')

        local = name in self._local_database.index()

        ex = ''
        assert local, ValueError(f'can not found the {name} in the '
                                            f'local {ex}databases'.format(name=name, ex=ex))
        if local:
            file = self._local_database.get_record(name)['o']
            if not file.closed:
                file.close()

            tmp = list(file.name)
            tmp.insert(file.name.find(name), 'stream_')
            streamfile_name = ''.join(tmp)
            try:
                os.remove(streamfile_name)
            except (FileNotFoundError, PermissionError):
                pass
            try:
                os.remove(file.name)
            except (PermissionError, FileNotFoundError):
                pass
            self._local_database.remove_record(name)

        gc.collect()

    def export(self, obj: Union[str, AudioMetadata, AudioWrap], file_path: str = './', format: FileFormat = FileFormat.UNKNOWN, quality: float = 0.5, bitrate: int = 128):
        '''
        Exports a record to a file in WAV, MP3, FLAC, or VORBIS format. The output format can be specified either through the `format` 
        argument or derived from the file extension in the `file_path`. If a file extension ('.wav', '.mp3', '.flac', or '.ogg') is 
        included in `file_path`, it takes precedence over the `format` argument. If no extension is provided, the 
        `format` argument is used, defaulting to WAV if set to FileFormat.UNKNOWN. The exported file is saved at the 
        specified `file_path`.

        :param obj: Record to export (str, AudioMetadata, or AudioWrap).
                    - str: Path to a file to be loaded and exported.
                    - AudioMetadata: A metadata object containing audio data.
                    - AudioWrap: Objects that wrap or generate the audio data.
        :param file_path: Path to save the exported file (default: './').
                        - A new filename can be specified at the end of the path.
                        - If a valid file extension ('.wav', '.mp3', '.flac', or '.ogg') is provided, it determines the output format, overriding the `format` argument.
                        - If no extension is included and the path is set to './', the name of the record is used.
        :param format: Output format (FileFormat.WAV, FileFormat.MP3, FileFormat.FLAC, or FileFormat.VORBIS). Defaults to FileFormat.UNKNOWN, 
                    which results in WAV being chosen unless a valid extension is provided in `file_path`.
        :param quality: Quality setting for encoding (default: 0.5).
                        - For WAV: Ignored
                        - For MP3: Converted to scale 0-9 (0 highest, 9 lowest)
                        - For FLAC: Converted to scale 0-8 (0 fastest/lowest, 8 slowest/highest)
                        - For VORBIS: Used directly (0.0 lowest, 1.0 highest)
        :param bitrate: Bitrate for MP3 encoding in kbps (default: 128). Only used if the format is MP3.
        :return: None

        Raises:

        - TypeError: Raised if `obj` is not one of the expected types (str, AudioMetadata, or AudioWrap).
        - ValueError: Raised if an unsupported format is provided.
        '''

        file_path = file_path.strip()
        if isinstance(obj, str): 
            record = self.load(obj, series=True)
        elif isinstance(obj, AudioWrap):
            record = obj.get_data()
        elif isinstance(obj, AudioMetadata):
            pass
        else:
            raise TypeError('please control the type of record')
        assert 0 <= quality <= 1, ValueError('Invalid quality factor')

        p0 = max(file_path.rfind('\\'), file_path.rfind('/'))
        p1 = file_path.rfind('.')
        if p0 < 0:
            p0 = 0
        if p1 <= 0:
            p1 = None
        
        if (not file_path) or (file_path == "./") or (file_path == ".") or (file_path == "/"):
            name = record.name
        else:
            name = file_path[p0 : p1]
        
        name_format = None
        if p1 is not None:
            name_format = file_path[p1 + 1:].lower()
            name_format = FileFormat.WAV if name_format == 'wav' else \
                        FileFormat.MP3 if name_format == 'mp3' else \
                        FileFormat.FLAC if name_format == 'flac' else \
                        FileFormat.VORBIS if name_format == 'ogg' else None

        format = name_format if name_format is not None else FileFormat.WAV if format == FileFormat.UNKNOWN else format
        supported = {FileFormat.WAV: 'wav', FileFormat.MP3: 'mp3', FileFormat.FLAC: 'flac', FileFormat.VORBIS: 'ogg'}
        if format not in supported:
            raise ValueError("Format must be either 'wav', 'mp3', 'flac', or 'ogg'")
        
        name += f'.{supported[format]}'
        if p0:
            file_path = file_path[0: p0 + 1] + name
        else:
            file_path = name
        
        file = record.o
        file_pos = file.tell()
        data = file.read()
        file.seek(file_pos, 0)
        
        if format == FileFormat.WAV:
            codec.encode_wav_file(
                file_path,
                data,
                record.sampleFormat,
                record.nchannels,
                record.frameRate,
            )
        elif format == FileFormat.MP3:
            mp3_quality = int(quality * 9)  # Convert 0-1 to 0-9 scale
            codec.encode_mp3_file(
                file_path,
                data,
                record.sampleFormat,
                record.nchannels,
                record.frameRate,
                bitrate,
                mp3_quality
            )
        elif format == FileFormat.FLAC:
            flac_compression = int(quality * 8)  # Convert 0-1 to 0-8 scale
            codec.encode_flac_file(
                file_path,
                data,
                record.sampleFormat,
                record.nchannels,
                record.frameRate,
                flac_compression
            )
        elif format == FileFormat.VORBIS:
            codec.encode_vorbis_file(
                file_path,
                data,
                record.sampleFormat,
                record.nchannels,
                record.frameRate,
                quality
            )


    def get_record_names(self) -> list:
        '''
        Returns a list of record names in the local database.
        '''
        return list(self._local_database.index())



    def get_nperseg(self):
        '''
        Returns the number of segments per window.
        '''
        return self._nperseg

    def get_nchannels(self):
        '''
        Returns the number of audio channels.
        '''
        return self._nchannels

    def get_sample_rate(self):
        '''
        Returns the sample rate of the master instanse core processor.
        '''
        return self._sample_rate

    def stream(self, record: Union[str, AudioMetadata, AudioWrap],
               block_mode: bool=False,
               safe_load: bool=False,
               on_stop: callable=None,
               loop_mode: bool=False,
               use_cached_files=True,
               stream_mode:StreamMode = StreamMode.optimized) -> StreamControl:
        '''
        Streams a record with optional loop and safe load modes.

        Note:
        
            The audio data maintaining process has additional cached files to reduce dynamic memory usage and improve performance,
            meaning that, The audio data storage methods can have different execution times based on the cached files.

        Note:

            The recorder can only capture normal streams(Non-optimized streams)

        :param record: Record to stream (str, AudioMetadata, or AudioWrap).
        :param block_mode: Whether to block the stream (default: `False`).
        :param safe_load: Whether to safely load the record (default: `False`). 
         load an audio file and modify it according to the 'Master' attributes(like the frame rate, number oof channels, etc).
        :param on_stop: Callback for when the stream stops (default: `None`).
        :param loop_mode: Whether to enable loop mode (default: `False`).
        :param use_cached_files: Whether to use cached files (default: `True`).
        :param stream_mode: Streaming mode (default: `StreamMode.optimized`).

        :return: A StreamControl object
        '''

        assert self.is_started(), "instance not started, use start()"

        # loop mode dont workes in blocking mode
        cache_head_check_size = 20

        if isinstance(record, str):
            record = self.load(record, series=True)
        else:
            if isinstance(record, AudioWrap): 
                assert record.is_packed(), BufferError('The {} is not packed!'.format(record))
                record = record.get_data()

            elif isinstance(record, AudioMetadata):
                record = record.copy()
            else:
                raise TypeError('please control the type of record')


        if record.nchannels > self._nchannels:
            raise ImportError('number of channel for the {name}({ch0})'
                              ' is not same as object channels({ch1})'.format(name=record.name,
                                                                              ch0=record.nchannels,
                                                                              ch1=self._nchannels))

        elif not self._sample_rate == record.frameRate:
            warnings.warn('Warning, frame rate must be same')

        assert  type(record.o) is BufferedRandom, TypeError('The record object is not standard')
        file = record.o
        file_pos = file.tell()
        tmp = list(file.name)
        tmp.insert(file.name.find(record.name), 'stream_')
        streamfile_name = ''.join(tmp)
        try:
            if use_cached_files:
                record_size = os.path.getsize(file.name)
                streamfile_size = os.path.getsize(streamfile_name)
                file.seek(0, 0)
                pre_head = file.read(Master.CACHE_INFO_SIZE + cache_head_check_size)
                file.seek(file_pos, 0)

                if record_size == streamfile_size:
                    pass
                elif (record_size - record.size) == streamfile_size:
                    pass
                    file_pos = Master.CACHE_INFO_SIZE
                else:
                    raise FileNotFoundError

                streamfile = open(streamfile_name, 'rb+')

                post_head = streamfile.read(Master.CACHE_INFO_SIZE + cache_head_check_size)
                streamfile.seek(file_pos, 0)
                if not post_head == pre_head:
                    streamfile.close()
                    raise FileNotFoundError

                if safe_load:
                    record.o = streamfile.read()
                    record = self._sync_record(record)
                    file.seek(file_pos, 0)
                record.o = streamfile
            else:
                raise FileNotFoundError

        except FileNotFoundError:
            record.o = file.read()
            if safe_load:
                record = self._sync_record(record)
            file.seek(file_pos, 0)

            record.o = streamfile = (
                write_to_cached_file(record.size,
                            record.frameRate,
                            record.sampleFormat if record.sampleFormat else 0,
                            record.nchannels,
                            file_name=streamfile_name,
                            data=record.o,
                            after_seek=(Master.CACHE_INFO_SIZE, 0),
                            after_flush=True)
            )
            file_pos = Master.CACHE_INFO_SIZE

        if block_mode:
            data_size = self._nperseg * self._nchannels * self._sample_width
            self._main_stream.acquire()
            while 1:
                in_data = streamfile.read(data_size)
                if not in_data:
                    break
                in_data = np.frombuffer(in_data, self._sample_width_format_str).astype('f')
                in_data = map_channels(in_data, self._nchannels, self._nchannels)
                self._main_stream.put(in_data)  
            self._main_stream.release()
            self._main_stream.clear()
            streamfile.seek(file_pos, 0)
            if on_stop:
                on_stop()

        else:
            return self._stream_control(record, on_stop, loop_mode, stream_mode)

    def _stream_control(self, *args):
        return StreamControl(self, *args)

    def mute(self):
        '''
        Mutes the master main stream.
        '''
        self._master_mute_mode.set()


    def unmute(self):
        '''
        Unmutes the master main stream.
        '''
        assert self.is_started(), "instance not started, use start()"
        assert not self._exstream_mode.is_set(), "stream is busy"
        self._master_mute_mode.clear()
        self._main_stream.clear()  # Clear any stale data in the stream


    def is_muted(self):
        '''
        Checks if the audio stream is muted.
        '''
        assert self.is_started(), "instance not started, use start()"
        return self._master_mute_mode.is_set()

    def echo(self, *args: Union[str, AudioMetadata, AudioWrap],
             enable: bool=None, main_output_enable: bool=False):
        
        """
        Play audio or manage audio streaming output functionality.

        Provides flexible audio playback and echo control with multiple input types 
        and configuration options.

        Parameters:
        -----------
        
        *args : str or AudioMetadata or AudioWrap
            Audio sources to play:
            - File path (str)
            - AudioMetadata object
            - AudioWrap object
            Multiple sources can be passed simultaneously

        enable : bool, optional
            Controls real-time echo behavior when no specific audio is provided:
            - None (default): Toggle echo on/off
            - True: Enable echo
            - False: Disable echo

        main_output_enable : bool, default False
            Determine whether to maintain the main audio stream's output during playback.
            Helps prevent potential audio feedback.

        Behavior:
        ---------
        
        - With no arguments: Manages real-time echo state
        - With audio sources: Plays specified audio through default output
        - Fallback to web audio if system audio is unavailable
        - Supports multiple audio playbacks

        Note: 
        -----

        If system audio fails, it'll try playing through your web browser 
        using HTML5 audio (if you're in a supported environment like Jupyter).

        Examples:
        ---------

            >>> master = sudio.Master()
            
            >>> master.add('audio1.ogg')
            >>> master.add('audio2.ogg')

            >>> # Play multiple audio sources
            >>> master.echo('audio1', 'audio2.wav')
            
            >>> # Toggle real-time echo
            >>> master.echo(enable=True)
        
        """
        if not len(args):
            if enable is None:
                if self._echo_flag.is_set():
                    self._echo_flag.clear()
                else:
                    self._main_stream.clear()
                    self._echo_flag.set()
            elif enable:
                self._main_stream.clear()
                self._echo_flag.set()
            else:
                self._echo_flag.clear()
        else:
            for rec in args:
                self._echo(rec, main_output_enable)

    def _echo(
        self, 
        record:Union[str, AudioMetadata, AudioWrap],
        main_output_enable: bool=False
        ):
        """
        Internal method for audio playback with flexible input handling.

        Processes different audio input types and manages playback through 
        system or web audio output mechanisms.

        Parameters:
        -----------

        record : str or AudioMetadata or AudioWrap
            Audio source to play:
            - File path (str)
            - AudioMetadata object
            - AudioWrap object with packed data

        main_output_enable : bool, default False
            Control main output stream during playback.
            Helps manage potential audio feedback scenarios.

        Raises:
        -------
        
        ValueError
            If an unsupported input type is provided.
        RuntimeError
            When no compatible audio output method is available.

        Notes:
        ------

        - Automatically handles different audio input formats
        - Attempts system audio output with fallback to web audio
        - Manages echo flag state during playback
        - Validates output device availability
        """
        if isinstance(record, AudioWrap): 
                assert  record.is_packed(), 'record should be packed!'
                record_data = record.get_data()
        else:
            if isinstance(record, str):
                record_data = self.load(record, series=True)
            elif isinstance(record, AudioMetadata):
                pass
            else:
                ValueError('unknown type')

        file = record_data.o
        assert not file.closed, "cache file is inaccessible."
        # file_pos = file.tell()
        file.seek(0, 0)
        data = file.read()
        file.seek(0, 0)

        flg = False
        if not main_output_enable and self._echo_flag.is_set():
            flg = True
            self._echo_flag.clear()

        try:
            assert self._output_device_index, "No output devices were found."
            write_to_default_output(
                data,
                record_data.sampleFormat,
                record_data.nchannels,
                record_data.frameRate,
                self._output_device_index
            )
        except Exception as e:
            try:
                if WebAudioIO.is_web_audio_supported():
                    result = WebAudioIO.play_audio_data(
                        data, 
                        record_data.sampleFormat, 
                        record_data.nchannels, 
                        record_data.frameRate,
                        )
                    assert result
                    warnings.warn(f'{e}')
                else:
                    raise
            except:
                raise RuntimeError(f"{e}")
                
        if flg:
            self._echo_flag.set()

    def disable_echo(self):
        '''
        Disables the echo functionality.
        '''
        assert self.is_started(), "instance not started, use start()"
        self.echo(enable=False)


    def wrap(self, record: Union[str, AudioMetadata]):
        '''
        wraps a record as a `AudioWrap`.
        
        :param record: Record to wrap (str or AudioMetadata).
        '''
        return AudioWrap(self, record)

    def prune_cache(self):
        """
        Retrieve a list of unused or orphaned cache files in the audio data directory.

        Scans the audio data directory and identifies cache files that are no longer
        referenced in the local database, helping manage file system resources.

        Returns:
        --------
        list
            Absolute file paths of cache files not associated with any current audio record.

        Notes:
        ------
        - Compares existing files against local database records
        - Filters out currently used cache files
        - Useful for identifying potential cleanup candidates
        """

        path = []
        expath = []
        base_len = len(self._audio_data_directory)
        for i in self._local_database.index():
            record = self._local_database.get_record(i)
            expath.append(record.o.name[base_len:])
        path += [i for i in expath if i not in path]

        listdir = os.listdir(self._audio_data_directory)
        listdir = list([self._audio_data_directory + item for item in listdir if item.endswith(Master.BUFFER_TYPE)])
        for i in path:
            j = 0
            while j < len(listdir):
                if i in listdir[j]:
                    del(listdir[j])
                else:
                    j += 1

        return listdir

    def clean_cache(self, max_retries = 3, retry_delay = .1):
        '''
        The audio data maintaining process has additional cached files to reduce dynamic
        memory usage and improve performance, meaning that, The audio data storage methods
        can have different execution times based on the cached files.
        This function used to clean the audio cache by removing cached files.
        
        The function implements retry logic for handling permission errors and ensures
        files are properly deleted across different operating systems.
        '''
        cache = self.prune_cache()
        
        for file_path in cache:
            path = Path(file_path)
            if not path.exists():
                continue
                
            for attempt in range(max_retries):
                try:
                    try:
                        path.chmod(0o666)
                    except:
                        pass
                        
                    path.unlink(missing_ok=True)
                    break  # Success - exit retry loop
                    
                except PermissionError:
                    if attempt < max_retries - 1:
                        # Wait before retry to allow potential file locks to clear
                        time.sleep(retry_delay)
                        continue
                        
                except OSError as e:
                    # Handle other OS-specific errors
                    if attempt < max_retries - 1:
                        time.sleep(retry_delay)
                        continue
                    else:
                        break
        time.sleep(retry_delay)


    def _refresh(self, *args):
        """
        Resets and reconfigures the signal processing buffers when things change.

        Handles window length updates, regenerates window functions, and clears 
        internal buffers. Useful when you need to adjust processing parameters 
        on the fly without completely rebuilding the entire stream setup.

        Optionally takes arguments for specific refresh scenarios, but can also 
        do a standard refresh based on current config.
        """

        win_len = self._nperseg

        if len(args):
            pass
        # primary filter frequency change
        else:

            if self._window_type:
                self._nhop = self._nperseg - self._noverlap
                self._nhop = int(self._nhop / self._data_chunk * win_len)

        if self._window is not None and (not self._nperseg == win_len):
            if type(self._window_type) is str or \
                    type(self._window_type) is tuple or \
                    type(self._window_type) == float:
                self._window = scisig.get_window(self._window_type, win_len)
            else:
                raise RefreshError("can't refresh static window")
            self._nperseg = win_len

        self._main_stream.acquire()


        self._overlap_buffer = [np.zeros(win_len) for i in range(self._nchannels)]
        self._windowing_buffer = [[np.zeros(win_len), np.zeros(win_len)] for i in range(self._nchannels)]

        self._refresh_ev.set()

        self._main_stream.clear()
        self._main_stream.release()

    def is_started(self):
        '''
        Checks if the audio input and output streams are started.
        '''
        return self._audio_input_stream and self._audio_output_stream
    
    def get_window(self):
        """
        Retrieves the current window configuration.

        :return: dict or None: A dictionary containing window information if available, or None if not set.

         - 'type': str, the type of window.
         - 'window': window data.
         - 'overlap': int, the overlap value.
         - 'size': int, the size of the window.

        """
        if self._window_type:
            return {'type': self._window_type,
                    'window': self._window,
                    'overlap': self._noverlap,
                    'size': len(self._window)}
        else:
            return None

    def disable_std_input(self):
        """
        Disables standard input stream by acquiring the main stream's lock object.
        """
        # if not self._main_stream.locked():
        assert self.is_started(), "instance not started, use start()"
        self._main_stream.acquire()

    def enable_std_input(self):
        """
        Enables standard input stream by clearing the main stream's lock.
        """
        assert self.is_started(), "instance not started, use start()"
        if self._main_stream.locked():
            self._main_stream.clear()
            self._main_stream.release()


    def add_pipeline(
            self, 
            pip, 
            name=None, 
            process_type: PipelineProcessType=PipelineProcessType.MAIN, 
            channel=None
            ):
        """
        Adds a new processing pipeline.

        Parameters:
        -----------

        - pip (obj): Pipeline object or array of defined pipelines.
         
        Note:
         In **PipelineProcessType.MULTI_STREAM** process type, pip must be an array of defined pipelines.
         The size of the array must be the same as the number of input channels.
         
        - name (str): Indicates the name of the pipeline.
        - process_type (PipelineProcessType): Type of processing pipeline (default: `PipelineProcessType.MAIN`). it can be:
         - **PipelineProcessType.MAIN**: Processes input data and passes it to activated pipelines (if exist).
         - **PipelineProcessType.BRANCH**: Represents a branch pipeline with optional channel parameter.
         - **PipelineProcessType.MULTI_STREAM**: Represents a multi_stream pipeline mode. Requires an array of pipelines.
        - channel (obj): None or [0 to self.nchannel].
         The input data passed to the pipeline can be a NumPy array in
         (self.nchannel, 2[2 windows in 1 frame], self._nperseg) dimension [None]
         or mono (2, self._nperseg) dimension.
         In mono mode [self.nchannel = 1 or mono mode activated], channel must be None.
        
        Note:
        
         The pipeline must process data and return it to the core with the dimensions as same as the input.

        """
        stream_type = 'multithreading'

        stream = None

        if name is None:
            name = generate_timestamp_name()

        if process_type == PipelineProcessType.MAIN or process_type == PipelineProcessType.MULTI_STREAM:
            stream = Stream(self, pip, name, stream_type=stream_type, process_type=process_type)
            # n-dim main pipeline
            self.main_pipe_database.append(stream)
        elif process_type == PipelineProcessType.BRANCH:
            stream = Stream(self, pip, name, stream_type=stream_type, channel=channel, process_type=process_type)
            self.branch_pipe_database.append(stream)
                
        else:
            pass

        return stream


    def set_pipeline(self, stream: Union[str, Stream]):
        '''
        sets the main processing pipeline.
        '''

        assert self.is_started(), "instance not started, use start()"

        if type(stream) is str:
            name = stream
        else:
            name = stream.name

        try:
            # find the specified pipeline
            pipeline = next(obj for obj in self.main_pipe_database if obj.name == name)
            
            assert pipeline.process_type in [PipelineProcessType.MAIN, PipelineProcessType.MULTI_STREAM], \
                f"Pipeline {name} is not a MAIN or MULTI_STREAM type"

            try:
                self.disable_pipeline()
            except ValueError:
                pass
            
            self._main_stream.acquire()
            try:
                self._main_stream.set(pipeline)
                
                if pipeline.process_type == PipelineProcessType.MULTI_STREAM:
                    for i in pipeline.pip:
                        assert i.is_alive(), f'Error: Sub-pipeline in {name} is not Enabled!'
                        i.sync(self._master_sync)
                    
                    for i in pipeline.pip:
                        i.aasync()

                # self._main_stream.clear()

            finally:
                self._main_stream.release()

        except StopIteration:
            raise ValueError(f"Pipeline {name} not found in main_pipe_database")
        except Exception as e:
            print(f"Error setting pipeline {name}: {str(e)}")
            # Attempt to restore the normal stream in case of error
            self.disable_pipeline()
            raise

    def disable_pipeline(self):
        '''
        Disables the current processing pipeline.
        '''

        assert self.is_started(), "instance not started, use start()"
        if self._main_stream and hasattr(self._main_stream, 'pip'):
            self._main_stream.acquire()
            if isinstance(self._main_stream.pip, Pipeline):
                self._main_stream.pip.clear()
            self._main_stream.set(self._normal_stream)
            
            self._main_stream.clear()
            
            self._main_stream.release()
        else:
            raise ValueError("No pipeline is currently set")             

    def clear_pipeline(self):
        '''
        Clears all pipeline's data.
        '''

        assert self.is_started(), "instance not started, use start()"
        if self._main_stream:
            self._main_stream.clear()
        for pipeline in self.main_pipe_database:
            pipeline.clear()
        for pipeline in self.branch_pipe_database:
            pipeline.clear()

    def _single_channel_windowing(self, data):
        """
        Performs windowing on a single channel of data.

        Parameters:
        - data (np.ndarray): The input data to be windowed.

        Returns:
        - np.ndarray: The windowed data.
        """
        # Check if the data is mono or multi-channel
        if self._window_type:
            retval = single_channel_windowing(
                data,
                self._windowing_buffer, 
                self._window,
                self._nhop
                )
        else:
            retval = data.astype(np.float64)
        return retval
    
    def _multi_channel_windowing(self, data):
        """
        Performs windowing on multiple channels of data.

        Parameters:
        - data (np.ndarray): The input data to be windowed.

        Returns:
        - np.ndarray: The windowed data.
        """
        # Check if the data is mono or multi-channel
        if self._window_type:
            retval = multi_channel_windowing(
                data,
                self._windowing_buffer,
                self._window,
                self._nhop,
                self._nchannels
                )
        else:
            retval = data.astype(np.float64)
        return retval
    

    def _single_channel_overlap(self, data):
        """
        Performs overlap-add on a single channel of data.

        Parameters:
        - data (np.ndarray): The input data to be processed.

        Returns:
        - np.ndarray: The processed data.
        """
        retval = single_channel_overlap(
            data,
            self._overlap_buffer,
            self._nhop
            )

        return retval
    

    def _multi_channel_overlap(self, data):
        """
        Performs overlap-add on multiple channels of data.

        Parameters:
        - data (np.ndarray): The input data to be processed.

        Returns:
        - np.ndarray: The processed data.
        """
        retval = multi_channel_overlap(
            data,
            self._overlap_buffer,
            self._nhop,
            self._nchannels,
            )

        return retval
    

    def set_window(self,
               window: object = 'hann',
               noverlap: int = None,
               NOLA_check: bool = True):
        '''
        Configures the window function for audio processing.

        :param window: The window function (default: `'hann'`).
        :param noverlap: Number of overlapping segments (default: `None`).
        :param NOLA_check: Perform the NOLA check (default: `True`).
        '''
        assert self.is_started(), "instance not started, use start()"

        if self._window_type == window and self._noverlap == noverlap:
            return
        else:
            self._window_type = window
            if noverlap is None:
                noverlap = self._nperseg // 2

            if type(window) is str or \
                    type(window) is tuple or \
                    type(window) == float:
                window = scisig.get_window(window, self._nperseg)

            elif window:
                assert len(window) == self._nperseg, 'control size of window'
                if NOLA_check:
                    assert scisig.check_NOLA(window, self._nperseg, noverlap)

            elif self._window_type is None:
                window = None
                self._window_type = None

        self._window = window
        # refresh
        self._noverlap = noverlap
        self._nhop = self._nperseg - self._noverlap

        self._overlap_buffer = [np.zeros(self._nperseg) for i in range(self._nchannels)]
        self._windowing_buffer = [[np.zeros(self._nperseg), np.zeros(self._nperseg)] for i in range(self._nchannels)]
        self._main_stream.clear()

    def get_sample_format(self)->SampleFormat:
        '''
        Returns the sample format of the master instance.
        '''
        return self._sample_format_type

    def _get_sample_format(self):
        '''
        Returns the sample format of the master instance.
        '''
        return self._sample_format

    @staticmethod
    def get_default_input_device_info()-> AudioDeviceInfo:
        """
        Returns information about the default input audio device.

        :return AudioDeviceInfo
        """
        data = AudioStream.get_default_input_device()
        return data

    @staticmethod
    def get_default_output_device_info()-> AudioDeviceInfo:
        """
        Returns information about the default output audio device.

        :return AudioDeviceInfo
        """
        data = AudioStream.get_default_output_device()
        return data
    
    @staticmethod
    def get_device_count()-> int:
        """
        Returns the number of available audio devices.
        """
        data = AudioStream.get_device_count()
        return data
    
    @staticmethod
    def get_device_info_by_index(index: int)-> AudioDeviceInfo:
        """
        Returns information about a specific audio device by index.

        :param index: The index of the audio device (int).

        :return AudioDeviceInfo
        """
        data = AudioStream.get_device_info_by_index(int(index))
        return data
    
    @staticmethod
    def get_input_devices():
        """
        Returns a list of available input devices.
        """
        return AudioStream.get_input_devices()

    @staticmethod
    def get_output_devices():
        """
        Returns a list of available output devices.
        """
        return AudioStream.get_output_devices()
